#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import copy
import dataclasses
import unittest
from unittest import mock
from uuid import UUID

from parameterized import parameterized

from evo.common.data import EmptyResponse, Environment, HTTPHeaderDict, HTTPResponse, Page, ResourceMetadata
from evo.common.test_tools import BASE_URL, utc_datetime

SAMPLE_HEADER_DICT = {
    "Content-Type": "application/json",
    "Authorization": "Bearer 123",
    "Cookie": "session=123",
    "Set-Cookie": "session=123",
}
ALTERNATE_HEADER_VALUES = {
    "Content-Type": "application/xml",
    "Authorization": "Bearer 456",
    "Cookie": "session=456",
    "Set-Cookie": "session=456",
}


# Uncomment the following import statement to compare header values that are usually hidden.
# from evo.common.test_tools import TestHTTPHeaderDict as HTTPHeaderDict


class TestHTTPHeaderDict(unittest.TestCase):
    def test_init_empty(self) -> None:
        """Test that the HTTPHeaderDict is empty when initialized with no arguments."""
        headers = HTTPHeaderDict()
        self.assertEqual(len(headers), 0)

    def assert_matches_sample_dict(self, headers: HTTPHeaderDict) -> None:
        """Assert that the given headers match the sample dict."""
        self.assertEqual(len(SAMPLE_HEADER_DICT), len(headers))
        for key, value in SAMPLE_HEADER_DICT.items():
            self.assertEqual(value, headers[key])

    def test_init_mapping(self) -> None:
        """Test that the HTTPHeaderDict is initialized correctly with a mapping."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        self.assert_matches_sample_dict(headers)

    def test_init_sequence(self) -> None:
        """Test that the HTTPHeaderDict is initialized correctly with a sequence."""
        headers = HTTPHeaderDict(list(SAMPLE_HEADER_DICT.items()))
        self.assert_matches_sample_dict(headers)

    def test_init_kwargs(self) -> None:
        """Test that the HTTPHeaderDict is initialized correctly with keyword arguments."""
        headers = HTTPHeaderDict(**SAMPLE_HEADER_DICT)
        self.assert_matches_sample_dict(headers)

    def test_update_mapping(self) -> None:
        """Test that updating the headers with a mapping works correctly."""
        headers = HTTPHeaderDict()
        headers.update(SAMPLE_HEADER_DICT)
        self.assert_matches_sample_dict(headers)

    def test_update_sequence(self) -> None:
        """Test that updating the headers with a sequence works correctly."""
        headers = HTTPHeaderDict()
        headers.update(list(SAMPLE_HEADER_DICT.items()))
        self.assert_matches_sample_dict(headers)

    def test_update_kwargs(self) -> None:
        """Test that updating the headers with keyword arguments works correctly."""
        headers = HTTPHeaderDict()
        headers.update(**SAMPLE_HEADER_DICT)
        self.assert_matches_sample_dict(headers)

    def test_update_appends(self) -> None:
        """Test that updating an existing field appends the new value."""
        # RFC 7230 section 3.2.2: Field Order.
        # A recipient MAY combine multiple header fields with the same field
        # name into one "field-name: field-value" pair, without changing the
        # semantics of the message, by appending each subsequent field value to
        # the combined field value in order, separated by a comma.  The order
        # in which header fields with the same field name are received is
        # therefore significant to the interpretation of the combined field
        # value;
        # https://www.rfc-editor.org/rfc/rfc7230#section-3.2.2
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        headers.update({"Authorization": "Bearer 456"})
        self.assertEqual("Bearer 123,Bearer 456", headers["Authorization"])

    def test_update_set_cookie_overwrites(self) -> None:
        """Test that updating a "Set-Cookie" field overwrites the existing value."""
        # In practice, the "Set-Cookie" header field ([RFC6265]) often
        # appears multiple times in a response message and does not use the
        # list syntax, violating the above requirements on multiple header
        # fields with the same name.  Since it cannot be combined into a
        # single field-value, recipients ought to handle "Set-Cookie" as a
        # special case while processing header fields.
        # https://www.rfc-editor.org/rfc/rfc7230#section-3.2.2
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        headers.update({"Set-Cookie": "session=456"})
        self.assertEqual("session=456", headers["Set-Cookie"])

    @parameterized.expand(
        [(field, value, ALTERNATE_HEADER_VALUES[field]) for field, value in SAMPLE_HEADER_DICT.items()]
    )
    def test_setitem(self, field: str, value: str, alternate: str) -> None:
        """Test that setting an item works correctly."""
        headers = HTTPHeaderDict()
        headers[field] = value
        self.assertEqual(value, headers[field])

        with self.subTest("Updating an existing field"):
            headers[field] = alternate
            if field == "Set-Cookie":
                self.assertEqual(alternate, headers[field])
            else:
                self.assertEqual(f"{value},{alternate}", headers[field])

        headers = HTTPHeaderDict()  # Reset the headers.
        with self.subTest("Field names are case-insensitive"):
            headers[field.lower()] = value
            self.assertEqual(value, headers[field.upper()])

    @parameterized.expand(SAMPLE_HEADER_DICT.items())
    def test_getitem(self, field: str, value: str) -> None:
        """Test that getting an item works correctly."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        self.assertEqual(value, headers[field])

        with self.subTest("Field names are case-insensitive"):
            for key in (field.lower(), field.upper(), field.title()):
                self.assertEqual(value, headers[key])

    @parameterized.expand(SAMPLE_HEADER_DICT.items())
    def test_get(self, field: str, value: str) -> None:
        """Test that getting an item works correctly."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        self.assertEqual(value, headers.get(field))

        with self.subTest("Field names are case-insensitive"):
            for key in (field.lower(), field.upper(), field.title()):
                self.assertEqual(value, headers.get(key))

    def test_get_default(self) -> None:
        """Test that the default value is returned when a header is not found."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        self.assertNotIn("Non-Existent-Header", headers)
        self.assertIsNone(headers.get("Non-Existent-Header"))
        self.assertEqual("default", headers.get("Non-Existent-Header", "default"))

    def test_get_no_default(self) -> None:
        """Test that `None` is returned when a header is not found and no default is provided."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        self.assertNotIn("Non-Existent-Header", headers)
        self.assertIsNone(headers.get("Non-Existent-Header"))

    @parameterized.expand(SAMPLE_HEADER_DICT.items())
    def test_delitem(self, field: str, value: str) -> None:
        """Test that deleting an item works correctly."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        self.assertIn(field, headers)
        self.assertEqual(len(headers), len(SAMPLE_HEADER_DICT))

        del headers[field]
        self.assertNotIn(field, headers)
        self.assertEqual(len(headers), len(SAMPLE_HEADER_DICT) - 1)

        with self.subTest("Field names are case-insensitive"):
            for key in (field.lower(), field.upper(), field.title()):
                headers[key] = value
                self.assertIn(field, headers)
                del headers[key]
                self.assertNotIn(field, headers)

    def test_contains(self) -> None:
        """Test that the headers can be checked for containment."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        for key in SAMPLE_HEADER_DICT:
            self.assertIn(key, headers)

    def test_keys(self) -> None:
        """Test that the keys method returns the keys in lowercase."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        self.assertListEqual(list(SAMPLE_HEADER_DICT.keys()), list(headers.keys()))

    def test_values(self) -> None:
        """Test that the values method returns the correct values."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        self.assertListEqual(list(SAMPLE_HEADER_DICT.values()), list(headers.values()))

    def test_items(self) -> None:
        """Test that the items method returns the correct items."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        self.assertListEqual(list(SAMPLE_HEADER_DICT.items()), list(headers.items()))

    def test_iter(self) -> None:
        """Test that the header fieldnames can be iterated over."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        self.assertListEqual(list(SAMPLE_HEADER_DICT), list(headers))

    def test_clear(self) -> None:
        """Test that the headers can be cleared."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        self.assertEqual(len(headers), len(SAMPLE_HEADER_DICT))
        headers.clear()
        self.assertEqual(len(headers), 0)

    @parameterized.expand(SAMPLE_HEADER_DICT)
    def test_copy(self, field: str) -> None:
        """Test that the headers can be copied."""
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        copy = headers.copy()
        self.assertEqual(headers, copy)

        with self.subTest("Updating the copy does not affect the original headers"):
            copy[field] = "bleh"
            self.assertNotEqual(headers, copy)

        with self.subTest("Updating the original headers does not affect the copy"):
            copy = headers.copy()
            self.assertEqual(headers, copy)
            headers[field] = "bleh"
            self.assertNotEqual(headers, copy)

    @parameterized.expand(SAMPLE_HEADER_DICT.items())
    def test_repr(self, field: str, expected_value: str) -> None:
        """Test that the repr method returns the expected string.

        Sensitive information should be hidden.
        """
        headers = HTTPHeaderDict(SAMPLE_HEADER_DICT)
        repr_headers = repr(headers)
        self.assertIsInstance(repr_headers, str)

        self.assertEqual(expected_value, headers[field])
        eval_headers = eval(repr_headers, globals(), {"TestHTTPHeaderDict": HTTPHeaderDict})

        if field in ("Authorization", "Proxy-Authorization", "Cookie", "Set-Cookie"):
            expected_value = "*****"
        self.assertEqual(expected_value, eval_headers[field], "Sensitive information should be hidden.")
        self.assertIn(f"{field.title()!r}: {expected_value!r}", repr_headers)


SAMPLE_RESPONSE_DICT = {
    "status": 200,
    "data": b"Hello, World!",
    "reason": "OK",
    "headers": SAMPLE_HEADER_DICT.copy(),
}


class TestEmptyResponse(unittest.TestCase):
    def setUp(self) -> None:
        self.response = EmptyResponse(
            status=SAMPLE_RESPONSE_DICT["status"],
            reason=SAMPLE_RESPONSE_DICT["reason"],
            headers=HTTPHeaderDict(SAMPLE_RESPONSE_DICT["headers"]),
        )

    def test_init_default_reason(self) -> None:
        """Test that the response can be initialized without a reason."""
        response = EmptyResponse(
            status=SAMPLE_RESPONSE_DICT["status"],
            headers=HTTPHeaderDict(SAMPLE_RESPONSE_DICT["headers"]),
        )
        self.assertIsNone(response.reason)

    def test_init_default_headers(self) -> None:
        """Test that the response can be initialized without headers."""
        response = EmptyResponse(
            status=SAMPLE_RESPONSE_DICT["status"],
            reason=SAMPLE_RESPONSE_DICT["reason"],
        )
        self.assertEqual(HTTPHeaderDict(), response.headers)

    def test_getheaders(self) -> None:
        """Test that getheaders returns a copy of the headers."""
        headers = self.response.getheaders()
        self.assertIsInstance(headers, HTTPHeaderDict)
        self.assertEqual(self.response.headers, headers)
        self.assertEqual(HTTPHeaderDict(SAMPLE_RESPONSE_DICT["headers"]), headers)

        with self.subTest("Updating the headers does not affect the original response"):
            headers["Content-Type"] = "application/xml"
            self.assertNotEqual(headers, self.response.headers)

    def test_getheader(self) -> None:
        """Test that getheader returns the correct header value."""
        for field, value in SAMPLE_RESPONSE_DICT["headers"].items():
            self.assertEqual(value, self.response.getheader(field))

            with self.subTest("Field names are case-insensitive"):
                for key in (field.lower(), field.upper(), field.title()):
                    self.assertEqual(value, self.response.getheader(key))

    def test_getheader_default(self) -> None:
        """Test that getheader returns the default value when the header is not found."""
        self.assertNotIn("Non-Existent-Header", self.response.headers)
        self.assertIsNone(self.response.getheader("Non-Existent-Header"))
        self.assertEqual("default", self.response.getheader("Non-Existent-Header", "default"))

    def test_getheader_no_default(self) -> None:
        """Test that getheader returns `None` when the header is not found and no default is provided."""
        self.assertNotIn("Non-Existent-Header", self.response.headers)
        self.assertIsNone(self.response.getheader("Non-Existent-Header"))


class TestHTTPResponse(TestEmptyResponse):
    def setUp(self) -> None:
        self.response = HTTPResponse(
            status=SAMPLE_RESPONSE_DICT["status"],
            reason=SAMPLE_RESPONSE_DICT["reason"],
            headers=HTTPHeaderDict(SAMPLE_RESPONSE_DICT["headers"]),
            data=SAMPLE_RESPONSE_DICT["data"],
        )

    def test_init_default_reason(self) -> None:
        """Test that the response can be initialized without a reason."""
        response = HTTPResponse(
            status=SAMPLE_RESPONSE_DICT["status"],
            headers=HTTPHeaderDict(SAMPLE_RESPONSE_DICT["headers"]),
            data=SAMPLE_RESPONSE_DICT["data"],
        )
        self.assertIsNone(response.reason)

    def test_init_default_headers(self) -> None:
        """Test that the response can be initialized without headers."""
        response = HTTPResponse(
            status=SAMPLE_RESPONSE_DICT["status"],
            reason=SAMPLE_RESPONSE_DICT["reason"],
            data=SAMPLE_RESPONSE_DICT["data"],
        )
        self.assertEqual(HTTPHeaderDict(), response.headers)


OFFSET = 0
LIMIT = 10
TOTAL = 200

# Test with one item less than the limit, because the limit is not guaranteed to be the page size.
COUNT = LIMIT - 1


@dataclasses.dataclass(frozen=True, kw_only=True)
class SomeItem(ResourceMetadata):
    extras: list = dataclasses.field(default_factory=list)

    @property
    def url(self) -> str:
        return f"{BASE_URL}/items/{self.id}"

    def __deepcopy__(self, memo: object) -> SomeItem:
        return dataclasses.replace(self, extras=copy.deepcopy(self.extras, memo))


def get_items(count: int, offset: int = 0) -> list[SomeItem]:
    mock_env = mock.Mock(spec=Environment)
    return [
        SomeItem(environment=mock_env, id=UUID(int=i), name=f"item-{i + offset}", created_at=utc_datetime(2024))
        for i in range(count)
    ]


class TestPage(unittest.TestCase):
    def setUp(self) -> None:
        self.items = get_items(COUNT, OFFSET)
        self.page = Page(offset=OFFSET, limit=LIMIT, total=TOTAL, items=self.items)

    def test_offset(self) -> None:
        self.assertEqual(OFFSET, self.page.offset)
        with self.assertRaises(AttributeError):
            self.page.offset = 1

    def test_limit(self) -> None:
        self.assertEqual(LIMIT, self.page.limit)
        with self.assertRaises(AttributeError):
            self.page.limit = 1

    def test_total(self) -> None:
        self.assertEqual(TOTAL, self.page.total)
        with self.assertRaises(AttributeError):
            self.page.total = 1

    def test_items(self) -> None:
        self.assertEqual(self.items, self.page.items())

    def assert_deepcopied(self, left: list[SomeItem], right: list[SomeItem]) -> None:
        self.assertEqual(left, right)
        for left_item, right_item in zip(left, right):
            self.assertIsNot(left_item, right_item)
            self.assertEqual(left_item.extras, right_item.extras)
            right_item.extras.append("extra")
            self.assertNotEqual(left_item.extras, right_item.extras)

    def test_items_copies_items(self) -> None:
        self.assert_deepcopied(self.items, self.page.items())

    def test_getitem(self) -> None:
        for i in range(COUNT):
            self.assertEqual(self.items[i], self.page[i])

    def test_getitem_copies_items(self) -> None:
        left = [self.items[0]]
        right = [self.page[0]]
        self.assert_deepcopied(left, right)

    def test_getitem_slice(self) -> None:
        left = self.items[1:5]
        right = self.page[1:5]
        self.assertEqual(left, right)

    def test_getitem_slice_copies_items(self) -> None:
        left = self.items[1:5]
        right = self.page[1:5]
        self.assert_deepcopied(left, right)

    def test_len(self) -> None:
        self.assertEqual(COUNT, len(self.page))

    def test_size(self) -> None:
        self.assertEqual(COUNT, self.page.size)

    @parameterized.expand(
        [
            ("first page", 0),
            ("some page", 42),  # The meaning of life, the universe, and everything.
            ("last page", TOTAL - COUNT),
        ]
    )
    def test_next_offset(self, _label: str, offset: int) -> None:
        page = Page(offset=offset, limit=LIMIT, total=TOTAL, items=get_items(COUNT, offset))
        expected_next_offset = offset + COUNT
        self.assertEqual(expected_next_offset, page.next_offset)

    @parameterized.expand(
        [
            ("first page", 0, False),
            ("some page", 42, False),
            ("last page", TOTAL - COUNT, True),
        ]
    )
    def test_is_last(self, _label: str, offset: int, expect_last: bool) -> None:
        page = Page(offset=offset, limit=LIMIT, total=TOTAL, items=get_items(COUNT, offset))
        self.assertEqual(expect_last, page.is_last)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Geoscience Object API
=============


The Geoscience Object API enables technological integrations at the object level. It enables users to access and utilise their data across all products through a common and accessible data structure.

A Geoscience Object is a data structure that represents a concrete geological, geotechnical, or geophysical concept. Geoscience Objects can be referenced by their UUID or by a user-defined object path.

For more information on using the Geoscience Object API, see the [Geoscience Object API overview](/docs/guides/objects), or the API references here.


This code is generated from the OpenAPI specification for Geoscience Object API.
API version: 1.14.0
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod

from ..models import *  # noqa: F403

__all__ = ["ObjectsApi"]


class ObjectsApi:
    """API client for the Objects endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def delete_object_by_path(
        self,
        org_id: str,
        workspace_id: str,
        objects_path: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> EmptyResponse:
        """Soft delete a geoscience object by path

        Soft delete a geoscience object by path. This endpoint will mark the object as deleted but won't actually remove it from the database.

        :param org_id: The customer's organisation ID
            Format: `uuid`
            Example: `'b208a6c9-6881-4b97-b02d-acb5d81299bb'`
        :param workspace_id:
            Format: `uuid`
            Example: `'032806a8-dcd7-11ed-8d5c-00155d8f28b5'`
        :param objects_path: Objects path. This parameter was automatically generated from a wildcard path.
            Example: `'objects_path_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
            "objects_path": objects_path,
        }

        # Prepare the header parameters.
        _header_params = {}
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "204": EmptyResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.DELETE,
            resource_path="/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/objects/path/{objects_path}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def delete_objects_by_id(
        self,
        object_id: str,
        org_id: str,
        workspace_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> EmptyResponse:
        """Soft delete a geoscience object

        Delete a geoscience object by UUID. Deleting an object to this endpoint will soft delete the existing object with a corresponding UUID. Note that the value of the `uuid` field in the object must match the value in the URL path.

        :param object_id:
            Format: `uuid`
            Example: `'object_id_example'`
        :param org_id: The customer's organisation ID
            Format: `uuid`
            Example: `'b208a6c9-6881-4b97-b02d-acb5d81299bb'`
        :param workspace_id:
            Format: `uuid`
            Example: `'032806a8-dcd7-11ed-8d5c-00155d8f28b5'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "object_id": object_id,
            "org_id": org_id,
            "workspace_id": workspace_id,
        }

        # Prepare the header parameters.
        _header_params = {}
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "204": EmptyResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.DELETE,
            resource_path="/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/objects/{object_id}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_object(
        self,
        org_id: str,
        workspace_id: str,
        objects_path: str,
        version: str | None = None,
        include_versions: bool | None = None,
        if_none_match: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> GetObjectResponse:  # noqa: F405
        """Download a geoscience object

        Download a geoscience object from the provided path. If required, the object can be compressed, currently gzip compression is supported. To receive a compressed payload set header `Accept-Encoding: gzip`.

        :param org_id: The customer's organisation ID
            Format: `uuid`
            Example: `'b208a6c9-6881-4b97-b02d-acb5d81299bb'`
        :param workspace_id:
            Format: `uuid`
            Example: `'032806a8-dcd7-11ed-8d5c-00155d8f28b5'`
        :param objects_path: Objects path. This parameter was automatically generated from a wildcard path.
            Example: `'objects_path_example'`
        :param version: (optional) The (timestamp) version of the object to retrieve. Defaults to the latest version if not specified.
            Example: `'2022-08-24T05:43:04.8822793Z'`
        :param include_versions: (optional) Whether to return all versions of the object
            Example: `False`
        :param if_none_match: (optional) Optional header used to check if there are any more recent versions of an object than the one specified. Returns 304 (Not Modified) if there are no newer versions.
            Example: `'1706583776404684724'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
            "objects_path": objects_path,
        }

        # Prepare the query parameters.
        _query_params = {}
        if version is not None:
            _query_params["version"] = version
        if include_versions is not None:
            _query_params["include_versions"] = include_versions

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if if_none_match is not None:
            _header_params["If-None-Match"] = if_none_match
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": GetObjectResponse,  # noqa: F405
            "304": EmptyResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/objects/path/{objects_path}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_object_by_id(
        self,
        object_id: str,
        org_id: str,
        workspace_id: str,
        version: str | None = None,
        include_versions: bool | None = None,
        deleted: bool | None = None,
        if_none_match: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> GetObjectResponse:  # noqa: F405
        """Download a geoscience object by UUID

        Download a geoscience object with the provided UUID. If required, the object can be compressed, currently gzip compression is supported. To receive a compressed payload set header `Accept-Encoding: gzip`.

        :param object_id: UUID of a geoscience object.
            Format: `uuid`
            Example: `'355fa5a6-f37d-11ed-93c1-00155d19a71b'`
        :param org_id: The customer's organisation ID
            Format: `uuid`
            Example: `'b208a6c9-6881-4b97-b02d-acb5d81299bb'`
        :param workspace_id:
            Format: `uuid`
            Example: `'032806a8-dcd7-11ed-8d5c-00155d8f28b5'`
        :param version: (optional) The (timestamp) version of the object to retrieve. Defaults to the latest version if not specified.
            Example: `'2022-08-24T05:43:04.8822793Z'`
        :param include_versions: (optional) Whether to return all versions of the object
            Example: `False`
        :param deleted: (optional) When true, only objects that have been deleted will be returned
            Example: `True`
        :param if_none_match: (optional) Optional header used to check if there are any more recent versions of an object than the one specified. Returns 304 (Not Modified) if there are no newer versions.
            Example: `'1706583776404684724'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "object_id": object_id,
            "org_id": org_id,
            "workspace_id": workspace_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if version is not None:
            _query_params["version"] = version
        if include_versions is not None:
            _query_params["include_versions"] = include_versions
        if deleted is not None:
            _query_params["deleted"] = deleted

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if if_none_match is not None:
            _header_params["If-None-Match"] = if_none_match
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": GetObjectResponse,  # noqa: F405
            "304": EmptyResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/objects/{object_id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def list_object_version_ids(
        self,
        org_id: str,
        workspace_id: str,
        request_body: list[str],
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> list[LatestObjectVersionIdResponse]:  # noqa: F405
        """List the latest version IDs for geoscience objects

        Request to fetch the latest version IDs for a collection of geoscience objects (maximum of 500). The request body must only contain a list of UUIDs of the objects to check. The response will contain a list of objects with their UUIDs and the latest version ID. If the object from a given UUID does not exist or has been deleted, the request will be successful, but the version ID will be `null`.

        :param org_id: The customer's organisation ID
            Format: `uuid`
            Example: `'b208a6c9-6881-4b97-b02d-acb5d81299bb'`
        :param workspace_id:
            Format: `uuid`
            Example: `'032806a8-dcd7-11ed-8d5c-00155d8f28b5'`
        :param request_body:
            Example: `['request_body_example']`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {
            "request_body": "",
        }

        _response_types_map = {
            "200": list[LatestObjectVersionIdResponse],  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.PATCH,
            resource_path="/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/objects",
            path_params=_path_params,
            header_params=_header_params,
            body=request_body,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def list_objects(
        self,
        org_id: str,
        workspace_id: str,
        limit: int | None = None,
        offset: int | None = None,
        deleted: bool | None = None,
        order_by: str | None = None,
        author: list[str] | None = None,
        created_by: list[str] | None = None,
        created_at: list[str] | None = None,
        modified_by: list[str] | None = None,
        modified_at: list[str] | None = None,
        object_name: list[str] | None = None,
        schema_id: list[str] | None = None,
        deleted_by: list[str] | None = None,
        deleted_at: list[str] | None = None,
        geojson_bounding_box: list[str] | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ListObjectsResponse:  # noqa: F405
        """List geoscience objects

        List geoscience objects. The response will include all geoscience objects matching the given filters.The geoscience objects will be the latest version of the object. If there are no objects matching the given filters, an empty list is returned.

        :param org_id: The customer's organisation ID
            Format: `uuid`
            Example: `'b208a6c9-6881-4b97-b02d-acb5d81299bb'`
        :param workspace_id:
            Format: `uuid`
            Example: `'032806a8-dcd7-11ed-8d5c-00155d8f28b5'`
        :param limit: (optional) Max number of results per page.
            Example: `5000`
        :param offset: (optional) Number of results to skip before returning `limit` number of results.
            Example: `0`
        :param deleted: (optional) When true, only objects that have been deleted will be returned
            Example: `True`
        :param order_by: (optional) A comma separated list of fields to order by, where the default sort order is ascending. To specify the sort order, prefix the field name with either `asc:` or `desc:` for ascending or descending respectively. Field names can either be arbitrary, representing a field nested inside the data, or one of the following known sort fields: `author`, `created_at`, `created_by`, `deleted_at`, `modified_at`, `modified_by`, `object_name`
            Example: `'order_by=name,desc:created_by,asc:object.a.b.c'`
        :param author: (optional) (deprecated, use `created_by`) The profile ID that you want to filter by
            Example: `['00000000-0000-0000-0000-000000000000']`
        :param created_by: (optional) The UUID of the user that created an object
            Example: `['00000000-0000-0000-0000-000000000000']`
        :param created_at: (optional)  A date or dates (max 2) to filter objects by. Dates may contain operator prefixes, in the form `<operator>:<datetime>`. The following operators are available (`lt`=less than, `lte`=less than or equal to, `gt`=greater than, `gte`=greater than or equal to). If you omit the operator, then it is assumed the operator is 'equal to'. In this case you may only supply one date. The dates must also be in a valid ISO 8601 format. Dates may include a UTC offset. If the offset is omitted, the timezone is assumed to be UTC.
            Example: `['gte:2023-03-10T22:56:53Z']`
        :param modified_by: (optional) The UUID of the user that modified an object
            Example: `['00000000-0000-0000-0000-000000000000']`
        :param modified_at: (optional)  A date or dates (max 2) to filter objects by. Dates may contain operator prefixes, in the form `<operator>:<datetime>`. The following operators are available (`lt`=less than, `lte`=less than or equal to, `gt`=greater than, `gte`=greater than or equal to). If you omit the operator, then it is assumed the operator is 'equal to'. In this case you may only supply one date. The dates must also be in a valid ISO 8601 format. Dates may include a UTC offset. If the offset is omitted, the timezone is assumed to be UTC.
            Example: `['gte:2023-03-10T22:56:53Z']`
        :param object_name: (optional) The name of the object to filter on. For backwards compatibility, when no operators are supplied, this will perform a case sensitive prefix match of the object file name. The query `object_name=gold` will match an object with the name `goldcolumns.json`. Providing an operator will query over the entire path to and including the file name.
            Example: `['object.json']`
        :param schema_id: (optional) The type of schema you want to filter by.
            Example: `['/objects/variogram/0.1.0/variogram.schema.json']`
        :param deleted_by: (optional) The UUID of the user that deleted an object
            Example: `['00000000-0000-0000-0000-000000000000']`
        :param deleted_at: (optional)  A date or dates (max 2) to filter objects by. Dates may contain operator prefixes, in the form `<operator>:<datetime>`. The following operators are available (`lt`=less than, `lte`=less than or equal to, `gt`=greater than, `gte`=greater than or equal to). If you omit the operator, then it is assumed the operator is 'equal to'. In this case you may only supply one date. The dates must also be in a valid ISO 8601 format. Dates may include a UTC offset. If the offset is omitted, the timezone is assumed to be UTC.
            Example: `['gte:2023-03-10T22:56:53Z']`
        :param geojson_bounding_box: (optional) The coordinates of the data to spatially search on. The value may optionally include one of two spatial operators, `geowithin` or `geointersects`. If an operator is not defined, then `geointersects` will be used by default. The coordinates must be 5 pairs of longitude and latitude, representing a closed polygon. The first and last coordinates must be the same to close the polygon.
            Example: `['geointersects:(171.6,-44.5),(173.7,-44.5),(173.7,-42.9),(171.6,-42.9),(171.6,-44.5)']`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if limit is not None:
            _query_params["limit"] = limit
        if offset is not None:
            _query_params["offset"] = offset
        if deleted is not None:
            _query_params["deleted"] = deleted
        if order_by is not None:
            _query_params["order_by"] = order_by
        if author is not None:
            _query_params["author"] = author
        if created_by is not None:
            _query_params["created_by"] = created_by
        if created_at is not None:
            _query_params["created_at"] = created_at
        if modified_by is not None:
            _query_params["modified_by"] = modified_by
        if modified_at is not None:
            _query_params["modified_at"] = modified_at
        if object_name is not None:
            _query_params["object_name"] = object_name
        if schema_id is not None:
            _query_params["schema_id"] = schema_id
        if deleted_by is not None:
            _query_params["deleted_by"] = deleted_by
        if deleted_at is not None:
            _query_params["deleted_at"] = deleted_at
        if geojson_bounding_box is not None:
            _query_params["geojson_bounding_box"] = geojson_bounding_box

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {
            "author": "multi",
            "created_by": "multi",
            "created_at": "multi",
            "modified_by": "multi",
            "modified_at": "multi",
            "object_name": "multi",
            "schema_id": "multi",
            "deleted_by": "multi",
            "deleted_at": "multi",
            "geojson_bounding_box": "multi",
        }

        _response_types_map = {
            "200": ListObjectsResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/objects",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def list_objects_by_org(
        self,
        org_id: str,
        offset: int | None = None,
        limit: int | None = None,
        deleted: bool | None = None,
        permitted_workspaces_only: bool | None = None,
        order_by: str | None = None,
        created_by: list[str] | None = None,
        created_at: list[str] | None = None,
        modified_by: list[str] | None = None,
        modified_at: list[str] | None = None,
        object_name: list[str] | None = None,
        schema_id: list[str] | None = None,
        geojson_bounding_box: list[str] | None = None,
        deleted_by: list[str] | None = None,
        deleted_at: list[str] | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ListOrgObjectsResponse:  # noqa: F405
        """List geoscience objects by org

        The response will list geoscience objects for the provided organisation ID.

        :param org_id: The customer's organisation ID
            Format: `uuid`
            Example: `'b208a6c9-6881-4b97-b02d-acb5d81299bb'`
        :param offset: (optional) Number of results to skip before returning `limit` number of results.
            Example: `0`
        :param limit: (optional) Max number of results per page.
            Example: `5000`
        :param deleted: (optional) When true, only objects that have been deleted will be returned
            Example: `True`
        :param permitted_workspaces_only: (optional) Only return objects in workspaces the user is permitted to access.
            Example: `True`
        :param order_by: (optional) A comma separated list of fields to order by, where the default sort order is ascending. To specify the sort order, prefix the field name with either `asc:` or `desc:` for ascending or descending respectively. Field names can either be arbitrary, representing a field nested inside the data, or one of the following known sort fields: `created_at`, `created_by`, `modified_at`, `modified_by`, `object_name`, `deleted_at`
            Example: `'order_by=name,desc:created_by,asc:object.a.b.c'`
        :param created_by: (optional) The UUID of the user that created an object
            Example: `['00000000-0000-0000-0000-000000000000']`
        :param created_at: (optional)  A date or dates (max 2) to filter objects by. Dates may contain operator prefixes, in the form `<operator>:<datetime>`. The following operators are available (`lt`=less than, `lte`=less than or equal to, `gt`=greater than, `gte`=greater than or equal to). If you omit the operator, then it is assumed the operator is 'equal to'. In this case you may only supply one date. The dates must also be in a valid ISO 8601 format. Dates may include a UTC offset. If the offset is omitted, the timezone is assumed to be UTC.
            Example: `['gte:2023-03-10T22:56:53Z']`
        :param modified_by: (optional) The UUID of the user that modified an object
            Example: `['00000000-0000-0000-0000-000000000000']`
        :param modified_at: (optional)  A date or dates (max 2) to filter objects by. Dates may contain operator prefixes, in the form `<operator>:<datetime>`. The following operators are available (`lt`=less than, `lte`=less than or equal to, `gt`=greater than, `gte`=greater than or equal to). If you omit the operator, then it is assumed the operator is 'equal to'. In this case you may only supply one date. The dates must also be in a valid ISO 8601 format. Dates may include a UTC offset. If the offset is omitted, the timezone is assumed to be UTC.
            Example: `['gte:2023-03-10T22:56:53Z']`
        :param object_name: (optional) The name of the object to filter on. For backwards compatibility, when no operators are supplied, this will perform a case sensitive prefix match of the object file name. The query `object_name=gold` will match an object with the name `goldcolumns.json`. Providing an operator will query over the entire path to and including the file name.
            Example: `['object.json']`
        :param schema_id: (optional) The type of schema you want to filter by.
            Example: `['/objects/variogram/0.1.0/variogram.schema.json']`
        :param geojson_bounding_box: (optional) The coordinates of the data to spatially search on. The value may optionally include one of two spatial operators, `geowithin` or `geointersects`. If an operator is not defined, then `geointersects` will be used by default. The coordinates must be 5 pairs of longitude and latitude, representing a closed polygon. The first and last coordinates must be the same to close the polygon.
            Example: `['geointersects:(171.6,-44.5),(173.7,-44.5),(173.7,-42.9),(171.6,-42.9),(171.6,-44.5)']`
        :param deleted_by: (optional) The UUID of the user that deleted an object
            Example: `['00000000-0000-0000-0000-000000000000']`
        :param deleted_at: (optional)  A date or dates (max 2) to filter objects by. Dates may contain operator prefixes, in the form `<operator>:<datetime>`. The following operators are available (`lt`=less than, `lte`=less than or equal to, `gt`=greater than, `gte`=greater than or equal to). If you omit the operator, then it is assumed the operator is 'equal to'. In this case you may only supply one date. The dates must also be in a valid ISO 8601 format. Dates may include a UTC offset. If the offset is omitted, the timezone is assumed to be UTC.
            Example: `['gte:2023-03-10T22:56:53Z']`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if offset is not None:
            _query_params["offset"] = offset
        if limit is not None:
            _query_params["limit"] = limit
        if deleted is not None:
            _query_params["deleted"] = deleted
        if permitted_workspaces_only is not None:
            _query_params["permitted_workspaces_only"] = permitted_workspaces_only
        if order_by is not None:
            _query_params["order_by"] = order_by
        if created_by is not None:
            _query_params["created_by"] = created_by
        if created_at is not None:
            _query_params["created_at"] = created_at
        if modified_by is not None:
            _query_params["modified_by"] = modified_by
        if modified_at is not None:
            _query_params["modified_at"] = modified_at
        if object_name is not None:
            _query_params["object_name"] = object_name
        if schema_id is not None:
            _query_params["schema_id"] = schema_id
        if geojson_bounding_box is not None:
            _query_params["geojson_bounding_box"] = geojson_bounding_box
        if deleted_by is not None:
            _query_params["deleted_by"] = deleted_by
        if deleted_at is not None:
            _query_params["deleted_at"] = deleted_at

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {
            "created_by": "multi",
            "created_at": "multi",
            "modified_by": "multi",
            "modified_at": "multi",
            "object_name": "multi",
            "schema_id": "multi",
            "geojson_bounding_box": "multi",
            "deleted_by": "multi",
            "deleted_at": "multi",
        }

        _response_types_map = {
            "200": ListOrgObjectsResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/geoscience-object/orgs/{org_id}/objects",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def post_objects(
        self,
        org_id: str,
        workspace_id: str,
        objects_path: str,
        geoscience_object: GeoscienceObject,  # noqa: F405
        if_match: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> PostObjectResponse:  # noqa: F405
        """Create a geoscience object

        Create or update a geoscience object.  Upload a geoscience object based on the schema element. Geoscience objects can be stored hierarchically,  by providing the path to the object, including the object name, in the request URL.  A detailed list of the available object schemas can be found in  Seequent's [evo-schemas](https://github.com/seequentevo/evo-schemas) repository.   The submitted properties will be validated against the specified schema properties.  To create a new object, upload it to a unique path with the UUID field set to `null`. The response will contain a reflection of the uploaded object with a newly assigned object ID. This can be used for future requests to update the object.  If a version of the geoscience object already exists at the specified path a new version will be created. To move or rename an existing object, simply upload a new version of it (with the same UUID) to a new, unique path or file name.  Note that every upload of an object is considered a new version, even if no changes have been made.  If required, the object should be compressed, currently gzip compression is supported. Set header `Content-Encoding: gzip` if the content is compressed and `Accept-Encoding: gzip` to receive compressed content.

        :param org_id: The customer's organisation ID
            Format: `uuid`
            Example: `'b208a6c9-6881-4b97-b02d-acb5d81299bb'`
        :param workspace_id:
            Format: `uuid`
            Example: `'032806a8-dcd7-11ed-8d5c-00155d8f28b5'`
        :param objects_path: Objects path. This parameter was automatically generated from a wildcard path.
            Example: `'objects_path_example'`
        :param geoscience_object:
            Example: `endpoints.GeoscienceObject()`
        :param if_match: (optional) Optional header to apply an update _only if_ the specified object version is the most recent available on the server.
            Example: `'1706583776404684724'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "workspace_id": workspace_id,
            "objects_path": objects_path,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if if_match is not None:
            _header_params["If-Match"] = if_match
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "201": PostObjectResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/objects/path/{objects_path}",
            path_params=_path_params,
            header_params=_header_params,
            body=geoscience_object,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def update_objects_by_id(
        self,
        object_id: str,
        org_id: str,
        workspace_id: str,
        deleted: bool | None = None,
        if_match: str | None = None,
        update_geoscience_object: UpdateGeoscienceObject | None = None,  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> PostObjectResponse:  # noqa: F405
        """Update a geoscience object

        Update or restore a geoscience object by UUID. Uploading an object to this endpoint will update (replace, i.e. no partial updates) the existing object with a corresponding UUID. If the object was previously deleted, and the `deleted` parameter is set to `false`, this operation will attempt to restore the object. You cannot provide an object body when attempting to restore an object. Note that the value of the `uuid` field in the object must match the value in the URL path. New objects must be uploaded by path.

        :param object_id: UUID of a geoscience object.
            Format: `uuid`
            Example: `'355fa5a6-f37d-11ed-93c1-00155d19a71b'`
        :param org_id: The customer's organisation ID
            Format: `uuid`
            Example: `'b208a6c9-6881-4b97-b02d-acb5d81299bb'`
        :param workspace_id:
            Format: `uuid`
            Example: `'032806a8-dcd7-11ed-8d5c-00155d8f28b5'`
        :param deleted: (optional) When set to false, the operation will attempt to restore the object.
            Example: `True`
        :param if_match: (optional) Optional header to apply an update _only if_ the specified object version is the most recent available on the server.
            Example: `'1706583776404684724'`
        :param update_geoscience_object: (optional)
            Example: `endpoints.UpdateGeoscienceObject()`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "object_id": object_id,
            "org_id": org_id,
            "workspace_id": workspace_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if deleted is not None:
            _query_params["deleted"] = deleted

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if if_match is not None:
            _header_params["If-Match"] = if_match
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "201": PostObjectResponse,  # noqa: F405
            "204": EmptyResponse,
            "303": PostObjectResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/objects/{object_id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=update_geoscience_object,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import uuid
from datetime import datetime
from typing import Iterable
from unittest import mock

import pyarrow

from evo.blockmodels import BlockModelAPIClient
from evo.blockmodels.endpoints import models
from evo.blockmodels.endpoints.models import JobResponse, JobStatus
from evo.blockmodels.exceptions import CacheNotConfiguredException, JobFailedException, MissingColumnInTable
from evo.common import ServiceUser
from evo.common.data import HTTPHeaderDict, RequestMethod
from evo.common.test_tools import BASE_URL, MockResponse, TestWithConnector, TestWithStorage
from utils import JobPollingRequestHandler

BM_UUID = uuid.uuid4()
GOOSE_UUID = uuid.uuid4()
GOOSE_VERSION_ID = "2"
DATE = datetime(2021, 1, 1)
MODEL_USER = models.UserInfo(email="test@test.com", name="Test User", id=uuid.uuid4())
USER = ServiceUser.from_model(MODEL_USER)


def _mock_version(
    version_id: int, version_uuid: uuid.UUID, goose_version_id: str, bbox=None, columns: Iterable[models.Column] = ()
) -> models.Version:
    return models.Version(
        base_version_id=None if version_id == 1 else version_id - 1,
        bbox=bbox,
        bm_uuid=BM_UUID,
        comment="",
        created_at=DATE,
        created_by=MODEL_USER,
        geoscience_version_id=goose_version_id,
        mapping=models.Mapping(columns=list(columns)),
        parent_version_id=version_id - 1,
        version_id=version_id,
        version_uuid=version_uuid,
    )


UPDATE_RESULT = models.UpdateWithUrl(
    changes=models.UpdateDataLiteOutput(
        # We don't look at these values, so we can just set them to empty
        columns=models.UpdateColumnsLiteOutput(new=[], update=[], rename=[], delete=[])
    ),
    version_uuid=uuid.uuid4(),
    job_uuid=uuid.uuid4(),
    job_url=f"{BASE_URL}/jobs/{uuid.uuid4()}",
    upload_url=f"{BASE_URL}/upload/{uuid.uuid4()}",
)

NEW_DATA = pyarrow.table(
    {"i": [1, 2, 3], "j": [4, 5, 6], "k": [7, 8, 9], "col1": ["A", "B", "B"], "col2": [4.5, 5.3, 6.2]}
)

UPDATED_VERSION = _mock_version(
    2,
    uuid.uuid4(),
    "3",
    models.BBox(
        i_minmax=models.IntRange(min=1, max=3),
        j_minmax=models.IntRange(min=4, max=6),
        k_minmax=models.IntRange(min=7, max=9),
    ),
    columns=[
        models.Column(col_id=str(uuid.uuid4()), title="col1", data_type=models.DataType.Utf8),
        models.Column(col_id=str(uuid.uuid4()), title="col2", data_type=models.DataType.Float64),
    ],
)


class UpdateRequestHandler(JobPollingRequestHandler):
    def __init__(
        self,
        update_result: models.UpdateWithUrl | None = None,
        job_response: JobResponse | None = None,
        pending_request: int = 0,
    ) -> None:
        super().__init__(job_response, pending_request)
        self._update_result = update_result

    async def request(
        self,
        method: RequestMethod,
        url: str,
        headers: HTTPHeaderDict | None = None,
        post_params: list[tuple[str, str | bytes]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> MockResponse:
        match method:
            case RequestMethod.POST if url.endswith("/uploaded"):
                job_url, _ = url.rsplit("/", 1)
                return MockResponse(status_code=201, content=json.dumps({"job_url": job_url}))
            case RequestMethod.PATCH:
                return MockResponse(status_code=202, content=self._update_result.model_dump_json())
            case RequestMethod.GET:
                return self.job_poll()
            case _:
                return self.not_found()


class TestUpdateBlockModel(TestWithConnector, TestWithStorage):
    def setUp(self) -> None:
        TestWithConnector.setUp(self)
        TestWithStorage.setUp(self)
        self.bms_client = BlockModelAPIClient(connector=self.connector, environment=self.environment, cache=self.cache)
        self.bms_client_without_cache = BlockModelAPIClient(connector=self.connector, environment=self.environment)

    @property
    def base_path(self) -> str:
        return f"blockmodel/orgs/{self.environment.org_id}/workspaces/{self.environment.workspace_id}"

    async def test_add_new_columns(self) -> None:
        self.transport.set_request_handler(
            UpdateRequestHandler(
                update_result=UPDATE_RESULT,
                job_response=JobResponse(
                    job_status=JobStatus.COMPLETE,
                    payload=UPDATED_VERSION,
                ),
            )
        )
        with (
            mock.patch("evo.common.io.upload.StorageDestination") as mock_destination,
        ):
            mock_destination.upload_file = mock.AsyncMock()
            version = await self.bms_client.add_new_columns(
                BM_UUID,
                NEW_DATA,
                units={"col2": "g/t"},
            )
            mock_destination.upload_file.assert_called_once()

            # Assert that the correct columns are part of the update
            expected_update_body = models.UpdateDataLiteInput(
                columns=models.UpdateColumnsLiteInput(
                    new=[
                        models.ColumnLite(
                            title="col1",
                            data_type=models.DataType.Utf8,
                            unit_id=None,
                        ),
                        models.ColumnLite(
                            title="col2",
                            data_type=models.DataType.Float64,
                            unit_id="g/t",
                        ),
                    ],
                    update=[],
                    rename=[],
                    delete=[],
                ),
                update_type=models.UpdateType.replace,
            )
            self.transport.assert_any_request_made(
                method=RequestMethod.PATCH,
                path=f"{self.base_path}/block-models/{BM_UUID}/blocks",
                body=expected_update_body.model_dump(mode="json", exclude_unset=True),
                headers={
                    "Authorization": "Bearer <not-a-real-token>",
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                },
            )
        self.assertEqual(version.bm_uuid, BM_UUID)
        self.assertEqual(version.version_id, 2)
        self.assertEqual(version.version_uuid, UPDATED_VERSION.version_uuid)
        self.assertEqual(version.parent_version_id, 1)
        self.assertEqual(version.base_version_id, 1)
        self.assertEqual(version.geoscience_version_id, "3")
        self.assertEqual(version.bbox, UPDATED_VERSION.bbox)
        self.assertEqual(version.comment, "")
        self.assertEqual(version.created_at, DATE)
        self.assertEqual(version.created_by, USER)
        self.assertEqual(version.columns, UPDATED_VERSION.mapping.columns)

    async def test_add_new_columns_no_cache(self) -> None:
        with self.assertRaises(CacheNotConfiguredException):
            await self.bms_client_without_cache.add_new_columns(
                BM_UUID,
                NEW_DATA,
            )

    async def test_add_new_columns_job_failed(self) -> None:
        self.transport.set_request_handler(
            UpdateRequestHandler(
                update_result=UPDATE_RESULT,
                job_response=JobResponse(
                    job_status=JobStatus.FAILED,
                    payload=models.JobErrorPayload(
                        detail="Update Job failed",
                        status=500,
                        title="Update Job failed",
                        type="https://seequent.com/error-codes/block-model-service/job/internal-error",
                    ),
                ),
            )
        )
        with (
            mock.patch("evo.common.io.upload.StorageDestination") as mock_destination,
        ):
            mock_destination.upload_file = mock.AsyncMock()
            with self.assertRaises(JobFailedException):
                await self.bms_client.add_new_columns(
                    BM_UUID,
                    NEW_DATA,
                )
            mock_destination.upload_file.assert_called_once()

    async def test_update_columns(self) -> None:
        self.transport.set_request_handler(
            UpdateRequestHandler(
                update_result=UPDATE_RESULT,
                job_response=JobResponse(
                    job_status=JobStatus.COMPLETE,
                    payload=UPDATED_VERSION,
                ),
            )
        )
        with (
            mock.patch("evo.common.io.upload.StorageDestination") as mock_destination,
        ):
            mock_destination.upload_file = mock.AsyncMock()
            version = await self.bms_client.update_block_model_columns(
                BM_UUID,
                NEW_DATA,
                new_columns=["col2"],
                update_columns={"col1"},
                delete_columns={"col3"},
                units={"col2": "g/t"},
            )
            mock_destination.upload_file.assert_called_once()

            # Assert that the correct columns are part of the update
            expected_update_body = models.UpdateDataLiteInput(
                columns=models.UpdateColumnsLiteInput(
                    new=[
                        models.ColumnLite(
                            title="col2",
                            data_type=models.DataType.Float64,
                            unit_id="g/t",
                        ),
                    ],
                    update=["col1"],
                    rename=[],
                    delete=["col3"],
                ),
                update_type=models.UpdateType.replace,
            )
            self.transport.assert_any_request_made(
                method=RequestMethod.PATCH,
                path=f"{self.base_path}/block-models/{BM_UUID}/blocks",
                body=expected_update_body.model_dump(mode="json", exclude_unset=True),
                headers={
                    "Authorization": "Bearer <not-a-real-token>",
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                },
            )
        self.assertEqual(version.bm_uuid, BM_UUID)
        self.assertEqual(version.version_id, 2)
        self.assertEqual(version.version_uuid, UPDATED_VERSION.version_uuid)
        self.assertEqual(version.parent_version_id, 1)
        self.assertEqual(version.base_version_id, 1)
        self.assertEqual(version.geoscience_version_id, "3")
        self.assertEqual(version.bbox, UPDATED_VERSION.bbox)
        self.assertEqual(version.comment, "")
        self.assertEqual(version.created_at, DATE)
        self.assertEqual(version.created_by, USER)
        self.assertEqual(version.columns, UPDATED_VERSION.mapping.columns)

    async def test_update_columns_no_cache(self) -> None:
        with self.assertRaises(CacheNotConfiguredException):
            await self.bms_client_without_cache.update_block_model_columns(
                BM_UUID,
                NEW_DATA,
                new_columns=["col2"],
            )

    async def test_update_columns_missing_in_data(self) -> None:
        with self.assertRaises(MissingColumnInTable):
            await self.bms_client.update_block_model_columns(
                BM_UUID,
                NEW_DATA,
                new_columns=["non_existent_column"],
            )

    async def test_update_columns_job_failed(self) -> None:
        self.transport.set_request_handler(
            UpdateRequestHandler(
                update_result=UPDATE_RESULT,
                job_response=JobResponse(
                    job_status=JobStatus.FAILED,
                    payload=models.JobErrorPayload(
                        detail="Update Job failed",
                        status=500,
                        title="Update Job failed",
                        type="https://seequent.com/error-codes/block-model-service/job/internal-error",
                    ),
                ),
            )
        )
        with (
            mock.patch("evo.common.io.upload.StorageDestination") as mock_destination,
        ):
            mock_destination.upload_file = mock.AsyncMock()
            with self.assertRaises(JobFailedException):
                await self.bms_client.update_block_model_columns(
                    BM_UUID,
                    NEW_DATA,
                    new_columns=["col2"],
                )
            mock_destination.upload_file.assert_called_once()



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

from parameterized import parameterized

from evo.oauth import EvoScopes, OAuthScopes

EXPECTED_SCOPE_NAMES = {
    EvoScopes.offline_access: "offline_access",
    EvoScopes.evo_discovery: "evo.discovery",
    EvoScopes.evo_workspace: "evo.workspace",
    EvoScopes.evo_blocksync: "evo.blocksync",
    EvoScopes.evo_object: "evo.object",
    EvoScopes.evo_file: "evo.file",
    EvoScopes.evo_audit: "evo.audit",
}


class TestEvoScopes(unittest.TestCase):
    @parameterized.expand(
        [
            (EvoScopes.offline_access,),
            (EvoScopes.evo_discovery,),
            (EvoScopes.evo_workspace,),
            (EvoScopes.evo_blocksync,),
            (EvoScopes.evo_object,),
            (EvoScopes.evo_file,),
            (EvoScopes.evo_audit,),
        ],
        name_func=lambda func,
        index,
        params: f"{func.__name__}_{index}_{parameterized.to_safe_name(str(params[0][0]))}",
    )
    def test_single_scope_string(self, scope: EvoScopes) -> None:
        self.assertIn(scope, EXPECTED_SCOPE_NAMES)
        expected_string = EXPECTED_SCOPE_NAMES[scope]
        self.assertEqual(expected_string, str(scope))
        self.assertEqual(1, len(scope.members), "Expected a single scope")

    @classmethod
    def expand_scopes(cls, scopes: tuple[EvoScopes, ...]) -> list[EvoScopes]:
        """Expand a tuple of scopes into a list of individual scopes."""
        all_scopes = set()
        for scope in scopes:
            all_scopes.update(scope.members)

        return sorted(all_scopes)

    @parameterized.expand(
        [
            (
                EvoScopes.default,
                (
                    EvoScopes.evo_discovery,
                    EvoScopes.evo_workspace,
                ),
            ),
            (
                EvoScopes.all_evo,
                (
                    EvoScopes.default,
                    EvoScopes.evo_blocksync,
                    EvoScopes.evo_object,
                    EvoScopes.evo_file,
                ),
            ),
        ],
        name_func=lambda f, n, p: f"{f.__name__}_{n}_{parameterized.to_safe_name(p[0][0].name)}",
    )
    def test_multiple_scope_string(self, scopes: EvoScopes, expected_scopes: tuple[EvoScopes, ...]) -> None:
        expected_string = " ".join([EXPECTED_SCOPE_NAMES[scope] for scope in self.expand_scopes(expected_scopes)])
        self.assertEqual(expected_string, str(scopes))

        for scope in expected_scopes:
            self.assertIn(scope, scopes)

        self.assertEqual(len(self.expand_scopes(expected_scopes)), len(scopes.members))

    def test_oauth_scope_alias(self) -> None:
        """Test that the OAuthScopes alias points to EvoScopes for backwards compatibility."""
        self.assertIs(EvoScopes, OAuthScopes)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from typing import TypeVar
from uuid import UUID

from ._types import DataType
from .endpoints import models
from .endpoints.models import JobResponse
from .exceptions import UnknownJobPayload

T = TypeVar("T")


def extract_payload(job_id: UUID, job_response: JobResponse, payload_type: type[T]) -> T:
    payload = job_response.payload
    if isinstance(payload, payload_type):
        return payload
    else:
        raise UnknownJobPayload(
            job_id, f"Expected {payload_type.__name__} for job payload, got {type(payload).__name__} instead"
        )


try:
    import pyarrow
except ImportError:
    data_type_mapping = {}
else:
    data_type_mapping = {
        pyarrow.bool_(): models.DataType.Boolean,
        pyarrow.int8(): models.DataType.Int8,
        pyarrow.int16(): models.DataType.Int16,
        pyarrow.int32(): models.DataType.Int32,
        pyarrow.int64(): models.DataType.Int64,
        pyarrow.uint8(): models.DataType.UInt8,
        pyarrow.uint16(): models.DataType.UInt16,
        pyarrow.uint32(): models.DataType.UInt32,
        pyarrow.uint64(): models.DataType.UInt64,
        pyarrow.float16(): models.DataType.Float16,
        pyarrow.float32(): models.DataType.Float32,
        pyarrow.float64(): models.DataType.Float64,
        pyarrow.utf8(): models.DataType.Utf8,
        pyarrow.date32(): models.DataType.Date32,
        pyarrow.timestamp("us", tz="UTC"): models.DataType.Timestamp,
    }


def convert_dtype(data_type: DataType) -> models.DataType:
    try:
        return data_type_mapping[data_type]
    except KeyError:
        raise ValueError(f"Unsupported data type: {data_type}")



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
from unittest import mock

from parameterized import parameterized

from evo.common.interfaces import IFeedback
from evo.common.utils import NoFeedback, PartialFeedback, iter_with_fb


class TestFeedback(unittest.TestCase):
    def setUp(self) -> None:
        self.parent_fb = mock.Mock(spec=IFeedback)

    def test_partial_feedback_full_range(self) -> None:
        fb = PartialFeedback(self.parent_fb, 0, 1)

        for p in [0.0, 0.0001, 0.1, 0.4999, 0.5, 0.5001, 0.9, 0.9999, 1.0]:
            fb.progress(p)
            self.parent_fb.progress.assert_called_once_with(p, None)
            self.parent_fb.reset_mock()

    @parameterized.expand(
        [
            ("first ten percent", 0.0, 0.1, [0.0, 0.0, 0.01, 0.05, 0.05, 0.05, 0.09, 0.1, 0.1]),
            ("middle of the range", 0.4, 0.6, [0.4, 0.4, 0.42, 0.5, 0.5, 0.5, 0.58, 0.6, 0.6]),
            ("last ten percent", 0.9, 1.0, [0.9, 0.9, 0.91, 0.95, 0.95, 0.95, 0.99, 1.0, 1.0]),
            ("very small part", 0.002, 0.003, [0.002, 0.002, 0.0021, 0.0025, 0.0025, 0.0025, 0.0029, 0.003, 0.003]),
        ]
    )
    def test_partial_feedback_partial_range(
        self, _name: str, start: float, end: float, expected_values: list[float]
    ) -> None:
        fb = PartialFeedback(self.parent_fb, start, end)

        for p, expected_p in zip([0.0, 0.0001, 0.1, 0.4999, 0.5, 0.5001, 0.9, 0.9999, 1.0], expected_values):
            fb.progress(p)
            self.parent_fb.progress.assert_called_once_with(expected_p, None)
            self.parent_fb.reset_mock()

    def test_iter_with_fb(self) -> None:
        n_elements = 10
        elements = [object() for _ in range(n_elements)]
        parent_fb: IFeedback = self.parent_fb
        for i, (element, fb) in enumerate(iter_with_fb(elements, parent_fb)):
            self.assertIs(elements[i], element)
            fb_part = i / n_elements

            with self.subTest("iter_with_fb updates progress"):
                if i > 0:  # FB is updated after each action.
                    self.parent_fb.progress.assert_called_once_with(fb_part)

            self.parent_fb.reset_mock()
            fb_part = round(fb_part + 0.5 / n_elements, ndigits=4)

            with self.subTest("fb from iter_with_fb updates progress"):
                fb.progress(0.5, "message")
                self.parent_fb.progress.assert_called_once_with(fb_part, "message")

            self.parent_fb.reset_mock()

        # FB is updated after each action.
        self.parent_fb.progress.assert_called_once_with(1.0)

    def test_iter_with_no_fb(self) -> None:
        n_elements = 10
        elements = [object() for _ in range(n_elements)]
        for i, (element, fb) in enumerate(iter_with_fb(elements)):
            self.assertIs(elements[i], element)
            self.assertIs(NoFeedback, fb)

    def test_iter_with_fb_no_elements(self) -> None:
        for _ in iter_with_fb([], self.parent_fb):
            raise AssertionError("Unreachable")

        self.parent_fb.progress.assert_not_called()



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import unittest
from typing import Literal

from parameterized import parameterized

from evo.aio import AioTransport
from evo.common.data import HTTPHeaderDict, RequestMethod
from evo.common.test_tools import BASE_URL, AbstractTestRequestHandler, MockResponse, TestTransport, long_test
from evo.common.utils import BackoffIncremental, BackoffMethod
from evo.oauth import AccessToken, OAuthConnector, OAuthError

RESPONSE_200_WITH_TOKEN = MockResponse(
    status_code=200,
    content=AccessToken(
        token_type="Bearer",
        access_token="not-a-real-token",
    ).model_dump_json(exclude_unset=True),
    headers={"Content-Type": "application/json"},
    reason="OK",
)
RESPONSE_500 = MockResponse(
    status_code=500,
    content=json.dumps(
        {
            "title": "Internal Server Error",
            "detail": "An unexpected error occurred.",
        },
    ),
    headers={"Content-Type": "application/json"},
    reason="Internal Server Error",
)
RESPONSE_502 = MockResponse(502)
RESPONSE_504 = MockResponse(504)


class IteratingTestRequestHandler(AbstractTestRequestHandler):
    def __init__(self, responses: tuple[MockResponse, ...]) -> None:
        self._responses = iter(responses)

    async def request(
        self,
        method: RequestMethod,
        url: str,
        headers: HTTPHeaderDict | None = None,
        post_params: list[tuple[str, str | bytes]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> MockResponse:
        return next(self._responses)


class TestOAuthConnector(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.transport = TestTransport()

    @parameterized.expand(
        [
            ("authorize", "/connect/authorize"),
            ("token", "/connect/token"),
        ],
    )
    def test_endpoint(self, endpoint_type: Literal["authorize", "token"], expected: str) -> None:
        transport = AioTransport(user_agent="test")
        connector = OAuthConnector(transport, client_id="test")
        assert connector.endpoint(endpoint_type) == expected

    @long_test
    @parameterized.expand(
        [
            (
                3,
                BackoffIncremental(0.02),
                (
                    RESPONSE_504,
                    RESPONSE_504,
                    RESPONSE_200_WITH_TOKEN,
                ),
                3,
                None,
            ),
            (
                3,
                BackoffIncremental(0.02),
                (
                    RESPONSE_502,
                    RESPONSE_502,
                    RESPONSE_200_WITH_TOKEN,
                ),
                3,
                None,
            ),
            (
                5,
                BackoffIncremental(0.02),
                (
                    RESPONSE_504,
                    RESPONSE_504,
                    RESPONSE_200_WITH_TOKEN,
                ),
                3,
                None,
            ),
            (
                5,
                BackoffIncremental(0.02),
                (
                    RESPONSE_500,
                    RESPONSE_500,
                    RESPONSE_500,
                ),
                1,
                OAuthError,
            ),
        ]
    )
    async def test_fetch_token(
        self,
        max_attempts: int,
        backoff_method: BackoffMethod,
        responses: tuple[MockResponse, ...],
        number_of_requests_that_should_be_made: int,
        raises: type[Exception] | None = None,
    ) -> None:
        self.transport.set_request_handler(IteratingTestRequestHandler(responses))
        connector = OAuthConnector(
            self.transport,
            base_uri=BASE_URL,
            client_id="test",
            max_attempts=max_attempts,
            backoff_method=backoff_method,
        )
        if raises:
            with self.assertRaises(raises):
                async with connector:
                    await connector.fetch_token({}, AccessToken)
        else:
            async with connector:
                await connector.fetch_token({}, AccessToken)
        self.transport.assert_n_requests_made(number_of_requests_that_should_be_made)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
from io import BytesIO
from unittest import mock

from parameterized import parameterized

from evo.common.io import BytesDestination, BytesSource

from .common import TestISource


class TestBytesDestination(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.buffer = BytesIO()
        self.destination = BytesDestination(self.buffer)

    @parameterized.expand(
        [
            ("next byte", 0, b"\xaa", b"\xaa"),
            ("skip bytes", 3, b"\xaa", b"\x00\x00\x00\xaa"),
        ]
    )
    async def test_write_chunk(self, _name: str, offset: int, data: bytes, expected_result: bytes) -> None:
        await self.destination.write_chunk(offset, data)
        self.buffer.seek(0)
        actual_result = self.buffer.read()
        self.assertEqual(expected_result, actual_result)


class TestBytesSource(TestISource):
    def setup_source(self, test_data: bytes) -> None:
        self.buffer = BytesIO(test_data)
        self.source = BytesSource(self.buffer, len(test_data))

    @parameterized.expand(
        [
            ("with buffer at start", 0),
            ("with buffer at end", 5),
            ("with buffer in middle", 2),
        ]
    )
    async def get_size_uninitialized(self, _label: str, offset_start: int) -> None:
        expected_size = self.buffer.seek(0, 2)
        self.buffer.seek(offset_start)

        # Mock the buffer to ensure that it is not being used prematurely.
        mock_buffer = mock.Mock(wraps=self.buffer)
        source = BytesSource(mock_buffer)

        # Ensure that the size is not calculated during initialization.
        mock_buffer.seek.assert_not_called()
        mock_buffer.tell.assert_not_called()

        # Ensure that the size is correctly calculated when `get_size` is called.
        actual = await source.get_size()
        self.assertEqual(expected_size, actual)
        mock_buffer.seek.assert_called_once_with(0, 2)

        # Ensure that the size is cached
        mock_buffer.seek.reset_mock()
        actual = await source.get_size()
        self.assertEqual(expected_size, actual)
        mock_buffer.seek.assert_not_called()


# Delete base test classes to prevent discovery by unittest.
del TestISource



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import random
import sys
from collections.abc import Iterator
from datetime import datetime, timezone
from io import BytesIO
from typing import Any

import jmespath
import numpy
import pyarrow as pa
import pyarrow.parquet as pq

from evo.objects.utils.tables import BaseTableFormat, _ColumnFormat


class NoImport:
    """Simple context manager to prevent one or more named modules from being imported."""

    def __init__(self, *names: str) -> None:
        """
        :param names: The names of the modules to prevent from being imported.
        """
        self._names = names
        self._unloaded_modules = {}

    def __enter__(self) -> None:
        for name in self._names:
            # If the module is already imported, save it and set to None.
            self._unloaded_modules[name] = sys.modules[name]
            # Set the module to None to prevent it from being re-imported.
            sys.modules[name] = None

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        # Restore the unloaded modules.
        for name, module in self._unloaded_modules.items():
            sys.modules[name] = module


class UnloadModule:
    """Simple context manager to unload one or more named modules on entry and restore on exit."""

    def __init__(self, *names: str) -> None:
        """
        :param names: The names of the modules to unload on entry and restore on exit.
        """
        self._names = names
        self._unloaded_modules = {}

    def _unload_module(self, name: str) -> None:
        if name in sys.modules:
            self._unloaded_modules[name] = sys.modules[name]
            del sys.modules[name]

        parent, *_ = name.rpartition(".")
        if parent:
            self._unload_module(parent)

    def __enter__(self) -> None:
        for name in self._names:
            self._unload_module(name)

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        # Restore the unloaded modules.
        for name, module in self._unloaded_modules.items():
            sys.modules[name] = module


def _generate_float64_data(n_samples: int) -> Iterator[float]:
    max_ = numpy.finfo("float64").max
    for _ in range(n_samples):
        yield max_ * random.uniform(-1.0, 1.0)


def _generate_int_data(int_type: str, n_samples: int) -> Iterator[int]:
    min_, max_ = numpy.iinfo(int_type).min, numpy.iinfo(int_type).max
    for _ in range(n_samples):
        yield random.randint(min_, max_)


def _generate_bool_data(n_samples: int) -> Iterator[bool]:
    for _ in range(n_samples):
        yield random.choice((True, False))


def _generate_string_data(n_samples: int) -> Iterator[str]:
    str_sample = "0123456789ABCDEF "
    for _ in range(n_samples):
        length = random.randint(10, 10000)
        yield "".join(random.choices(str_sample, k=length))


def _generate_timestamp_data(n_samples: int) -> Iterator[datetime]:
    min_ = datetime(1970, 1, 1, tzinfo=timezone.utc).timestamp()
    max_ = datetime(2038, 12, 31, 23, 59, 59, 999999, tzinfo=timezone.utc).timestamp()
    for _ in range(n_samples):
        yield datetime.utcfromtimestamp(random.uniform(min_, max_))


def _generate_data(format_id: str, n_samples: int) -> Iterator:
    match format_id:
        case "float64":
            yield from _generate_float64_data(n_samples)
        case "uint8" | "uint32" | "uint64" | "int32" | "int64" as int_type:
            yield from _generate_int_data(int_type, n_samples)
        case "bool":
            yield from _generate_bool_data(n_samples)
        case "string":
            yield from _generate_string_data(n_samples)
        case "timestamp":
            yield from _generate_timestamp_data(n_samples)
        case unknown_format:
            raise TypeError(f"Unsupported format '{unknown_format}'")


def _change_format(current_format: _ColumnFormat) -> _ColumnFormat:
    match current_format.id:
        case "float64":
            return _ColumnFormat("int64")
        case "uint8" | "uint32" | "uint64" | "int32" | "int64":
            return _ColumnFormat("float64")
        case "bool" | "timestamp":
            return _ColumnFormat("string")
        case "string":
            return _ColumnFormat("bool")
        case unknown_format:
            raise TypeError(f"Unsupported format '{unknown_format}'")


def get_sample_table(
    table_format: BaseTableFormat, n_rows: int, add_column: bool = False, change_types: bool = False
) -> pa.Table:
    column_formats = [column for column in table_format._columns]

    if add_column:
        column_formats.append(_ColumnFormat(column_formats[-1].type))

    if change_types:
        column_formats = [_change_format(column) for column in column_formats]

    if table_format._multi_dimensional:
        # Test multidimensional tables with an arbitrary number of columns. If the number of columns matches a more
        # specific GO type (one with a fixed number of columns), the more specific type would be instantiated.
        column_formats *= 20

    sample_schema = pa.schema(
        [pa.field(f"{column.id}[{i}]", column.type, nullable=False) for i, column in enumerate(column_formats)]
    )
    sample_data = [
        pa.array(_generate_data(column_format.id, n_rows), type=column_format.type, size=n_rows)
        for column_format in column_formats
    ]
    return pa.table(sample_data, names=sample_schema.names).cast(sample_schema)


def get_sample_table_and_bytes(table_format: BaseTableFormat, n_rows: int) -> tuple[pa.Table, bytes]:
    table = get_sample_table(table_format, n_rows)
    return table, write_table_to_bytes(table)


def write_table_to_bytes(table: pa.Table) -> bytes:
    memory = BytesIO()
    pq.write_table(table, where=memory, version="2.4", compression="gzip")
    return memory.getvalue()


# Support for assignment operations using JMESPath expressions.
# Could be moved to evo.jmespath in the future, if we want to expose this functionality outside of tests.
class _AssignmentTargetDictEntry:
    """Represents a dictionary entry that potentially can be assigned to."""

    def __init__(self, key: str, obj: dict):
        self.key = key
        self.obj = obj

    @property
    def value(self) -> Any:
        """Get the value at this dictionary entry, creating an empty dict if it doesn't exist."""
        return self.obj.setdefault(self.key, {})


class _AssignmentTargetListEntry:
    """Represents a list entry that potentially can be assigned to."""

    def __init__(self, index: int, obj: list):
        self.index = index
        self.obj = obj

    @property
    def value(self) -> Any:
        """Get the value at this list entry, or None if the index is out of range."""
        try:
            return self.obj[self.index]
        except IndexError:
            return None


class _AssignInterpreter(jmespath.visitor.Visitor):
    """A JMESPath visitor used for processing assignment operations.

    This only supports a subset of JMESPath expressions that can be used for assignment.

    This works by lazily evaluating field and index accesses, so that the last operation can be turned into an
    assignment. If another operation is encountered after a field or index access, the value is evaluated at that
    point.
    """

    def default_visit(self, node, *args, **kwargs):
        raise NotImplementedError(node["type"])

    @staticmethod
    def _evaluate_value(value):
        """Lazily evaluate the value if it's an assignment target."""
        if isinstance(value, (_AssignmentTargetDictEntry, _AssignmentTargetListEntry)):
            return value.value
        else:
            return value

    def visit_field(self, node, value):
        """Visit a field access node, i.e. foo.bar."""
        evaluated_value = self._evaluate_value(value)
        if not isinstance(evaluated_value, dict):
            return None
        return _AssignmentTargetDictEntry(node["value"], evaluated_value)

    def visit_index(self, node, value):
        """Visit an index access node, i.e. foo[0]."""
        evaluated_value = self._evaluate_value(value)
        if not isinstance(evaluated_value, list):
            return None
        return _AssignmentTargetListEntry(node["value"], evaluated_value)

    def _visit_sub_or_index_expression(self, node, value):
        """Visit a subexpression or index expression node, i.e. foo.bar.baz or a[0][1]."""
        result = value
        for node in node["children"]:
            result = self.visit(node, result)
        return result

    visit_subexpression = _visit_sub_or_index_expression
    visit_index_expression = _visit_sub_or_index_expression


def assign_property(obj: dict, expression: str, value: Any) -> None:
    """Assign a value to a property in a dictionary using a JMESPath expression.

     This only supports a subset of JMESPath expressions that can be used for assignment. In particular, only the following
    expression types are supported:
    - Field accesses (e.g. foo.bar)
    - Index accesses (e.g. foo[0])
    - Subexpressions combining the above (e.g. foo.bar[0].baz)
    If the expression is not in that form, a JMESPathError will be raised.

    Also, if the expression attempts to perform an invalid operation like:
    - Accessing a field on a non-object
    - Accessing an index on a non-array
    - Accessing an out-of-bounds index on an array
    then a JMESPathError will be raised.

    Accessing a non-existent field on an object will create an empty object at that field to allow for nested assignments.

    :param obj: The dictionary to assign the property to.
    :param expression: The JMESPath expression representing the property to assign to.
    :param value: The value to assign to the property.
    """
    parsed_expression = jmespath.compile(expression)
    interpreter = _AssignInterpreter()
    target = interpreter.visit(parsed_expression.parsed, obj)

    if isinstance(target, _AssignmentTargetDictEntry):
        target.obj[target.key] = value
    elif isinstance(target, _AssignmentTargetListEntry):
        target.obj[target.index] = value
    else:
        raise TypeError(f"Cannot assign to expression '{expression}'")



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
from datetime import datetime, timezone
from uuid import UUID

from parameterized import parameterized

from evo.common import Environment
from evo.workspaces import ServiceUser, UserPermission, Workspace, WorkspaceRole
from evo.workspaces.endpoints.models import RoleEnum, UserModel
from evo.workspaces.exceptions import UserPermissionTypeError


class TestWorkspaceRoles(unittest.TestCase):
    """Test cases for the workspace roles and permissions enums."""

    @parameterized.expand(
        [
            (UserPermission.read, UserPermission.write),
            (UserPermission.read, UserPermission.manage),
            (UserPermission.write, UserPermission.manage),
        ]
    )
    def test_user_permissions_unique(self, left: UserPermission, right: UserPermission) -> None:
        """Test that user permissions enums are unique."""
        self.assertNotEqual(left.value, right.value)
        masked = left & right
        self.assertEqual(masked, UserPermission(0))

    @parameterized.expand(
        [
            ("Viewer can read", WorkspaceRole.viewer, UserPermission.read),
            ("Editor can read", WorkspaceRole.editor, UserPermission.read),
            ("Editor can write", WorkspaceRole.editor, UserPermission.write),
            ("Editor is viewer", WorkspaceRole.editor, WorkspaceRole.viewer),
            ("Owner can read", WorkspaceRole.owner, UserPermission.read),
            ("Owner can write", WorkspaceRole.owner, UserPermission.write),
            ("Owner can manage", WorkspaceRole.owner, UserPermission.manage),
            ("Owner is viewer", WorkspaceRole.owner, WorkspaceRole.viewer),
            ("Owner is editor", WorkspaceRole.owner, WorkspaceRole.editor),
        ]
    )
    def test_workspace_role_permissions_include(
        self, _label: str, role: WorkspaceRole, permission: UserPermission | WorkspaceRole
    ) -> None:
        """Test that workspace roles contain the correct permissions."""
        self.assertIn(permission, role)

    @parameterized.expand(
        [
            ("Viewer cannot write", WorkspaceRole.viewer, UserPermission.write),
            ("Viewer cannot manage", WorkspaceRole.viewer, UserPermission.manage),
            ("Viewer is not an editor", WorkspaceRole.viewer, WorkspaceRole.editor),
            ("Viewer is not an owner", WorkspaceRole.viewer, WorkspaceRole.owner),
            ("Editor cannot manage", WorkspaceRole.editor, UserPermission.manage),
            ("Editor is not an owner", WorkspaceRole.editor, WorkspaceRole.owner),
        ]
    )
    def test_workspace_role_permissions_exclude(
        self, _label: str, role: WorkspaceRole, permission: UserPermission | WorkspaceRole
    ) -> None:
        """Test that workspace roles do not contain the incorrect permissions."""
        self.assertNotIn(permission, role)

    def test_user_permission_type_error(self) -> None:
        """Test that a user permission type error is raised when an invalid permission is used."""
        with self.assertRaises(UserPermissionTypeError):
            assert RoleEnum.viewer in WorkspaceRole.viewer

    @parameterized.expand(
        [
            ("Viewer is a viewer", RoleEnum.viewer, WorkspaceRole.viewer),
            ("Editor is an editor", RoleEnum.editor, WorkspaceRole.editor),
            ("Owner is an owner", RoleEnum.owner, WorkspaceRole.owner),
        ]
    )
    def test_endpoint_role_conversion(
        self, _label: str, endpoint_role: RoleEnum, workspace_role: WorkspaceRole
    ) -> None:
        """Test that the endpoint role enum can be converted to the workspace role enum."""
        interpreted_role = WorkspaceRole[str(endpoint_role.value)]
        self.assertEqual(workspace_role, interpreted_role)


class TestUser(unittest.TestCase):
    """Test cases for the service user dataclass."""

    def test_from_model(self) -> None:
        """Test that a service user can be created from a user model."""
        model = UserModel(id=UUID(int=0), name="Test User", email="test.user@unit.test")
        user = ServiceUser.from_model(model)
        self.assertEqual(model.id, user.id)
        self.assertEqual(model.name, user.name)
        self.assertEqual(model.email, user.email)

    def test_from_model_no_name(self) -> None:
        """Test that a service user can be created from a user model with no name."""
        model = UserModel(id=UUID(int=0), name=None, email="test.user@unit.test")
        user = ServiceUser.from_model(model)
        self.assertEqual(model.id, user.id)
        self.assertIsNone(user.name)
        self.assertEqual(model.email, user.email)

    def test_from_model_no_email(self) -> None:
        """Test that a service user can be created from a user model with no email."""
        model = UserModel(id=UUID(int=0), name="Test User", email=None)
        user = ServiceUser.from_model(model)
        self.assertEqual(model.id, user.id)
        self.assertEqual(model.name, user.name)
        self.assertIsNone(user.email)

    def test_from_model_no_name_or_email(self) -> None:
        """Test that a service user can be created from a user model with no name or email."""
        model = UserModel(id=UUID(int=0), name=None, email=None)
        user = ServiceUser.from_model(model)
        self.assertEqual(model.id, user.id)
        self.assertIsNone(user.name)
        self.assertIsNone(user.email)


class TestWorkspace(unittest.TestCase):
    """Test cases for the workspace dataclass."""

    def setUp(self) -> None:
        self.org_id = UUID(int=0)
        self.workspace_id = UUID(int=1)
        self.user_id = UUID(int=2)
        self.user = ServiceUser(id=self.user_id, name="Test User", email="test.user@unit.test")

    def get_workspace(self, role: WorkspaceRole = WorkspaceRole.owner) -> Workspace:
        return Workspace(
            id=self.workspace_id,
            display_name="Test Workspace",
            description="A test workspace",
            user_role=role,
            org_id=self.org_id,
            hub_url="https://unit.test/",
            created_at=datetime.fromtimestamp(0, tz=timezone.utc),
            created_by=self.user,
            updated_at=datetime.fromtimestamp(0, tz=timezone.utc),
            updated_by=self.user,
        )

    @parameterized.expand(
        [
            ("Viewer can read", WorkspaceRole.viewer, UserPermission.read),
            ("Editor can read", WorkspaceRole.editor, UserPermission.read),
            ("Editor can write", WorkspaceRole.editor, UserPermission.write),
            ("Editor is viewer", WorkspaceRole.editor, WorkspaceRole.viewer),
            ("Owner can read", WorkspaceRole.owner, UserPermission.read),
            ("Owner can write", WorkspaceRole.owner, UserPermission.write),
            ("Owner can manage", WorkspaceRole.owner, UserPermission.manage),
            ("Owner is viewer", WorkspaceRole.owner, WorkspaceRole.viewer),
            ("Owner is editor", WorkspaceRole.owner, WorkspaceRole.editor),
        ]
    )
    def test_user_has_permission(
        self, _label: str, role: WorkspaceRole, permission: UserPermission | WorkspaceRole
    ) -> None:
        """Test that the user has the correct permissions in the workspace."""
        workspace = self.get_workspace(role)
        self.assertTrue(workspace.user_has_permission(permission))

    def test_get_environment(self) -> None:
        """Test that the workspace can generate an environment object."""
        workspace = self.get_workspace()
        environment = workspace.get_environment()
        self.assertIsInstance(environment, Environment)
        self.assertEqual(self.org_id, environment.org_id)

    def test_hash(self) -> None:
        """Test that the workspace can be hashed."""
        workspace = self.get_workspace()
        self.assertEqual(hash(workspace.id), hash(workspace))



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Geoscience Object API
=============


The Geoscience Object API enables technological integrations at the object level. It enables users to access and utilise their data across all products through a common and accessible data structure.

A Geoscience Object is a data structure that represents a concrete geological, geotechnical, or geophysical concept. Geoscience Objects can be referenced by their UUID or by a user-defined object path.

For more information on using the Geoscience Object API, see the [Geoscience Object API overview](/docs/guides/objects), or the API references here.


This code is generated from the OpenAPI specification for Geoscience Object API.
API version: 1.14.0
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod

from ..models import *  # noqa: F403

__all__ = ["MetadataApi"]


class MetadataApi:
    """API client for the Metadata endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def update_metadata(
        self,
        object_id: str,
        org_id: str,
        workspace_id: str,
        metadata_update_body: MetadataUpdateBody,  # noqa: F405
        version_id: int | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> EmptyResponse:
        """Update metadata for an Object

        Update metadata for a geoscience object version. If a version is not specified, the latest version of the object will be used. Setting the metadata property to `null` will remove the metadata from the object version.

        :param object_id:
            Format: `uuid`
            Example: `'object_id_example'`
        :param org_id: The customer's organisation ID
            Format: `uuid`
            Example: `'b208a6c9-6881-4b97-b02d-acb5d81299bb'`
        :param workspace_id:
            Format: `uuid`
            Example: `'032806a8-dcd7-11ed-8d5c-00155d8f28b5'`
        :param metadata_update_body:
            Example: `endpoints.MetadataUpdateBody()`
        :param version_id: (optional) The version of the object to retrieve. Defaults to the latest version if not specified.
            Example: `1706583776404684724`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "object_id": object_id,
            "org_id": org_id,
            "workspace_id": workspace_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if version_id is not None:
            _query_params["version_id"] = version_id

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "204": EmptyResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.PATCH,
            resource_path="/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/objects/{object_id}/metadata",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=metadata_update_body,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import random
from uuid import UUID

from data import load_test_data
from evo.common.test_tools import TestWithConnector, TestWithDownloadHandler, TestWithUploadHandler
from evo.objects import ObjectDataDownload, ObjectDataUpload, ObjectMetadata

# The test data for these tests does need to be real parquet data, we just need enough content to test
# chunked upload and download.
TEST_DATA = random.randbytes(1024 * 1024 * 5)  # 5MB

OBJECT_ID = UUID(int=5)
VERSION_ID = "123456"
DATA_NAME = "0000000000000000000000000000000000000000000000000000000000000001"
INITIAL_URL = "https://unit.test/initial/url"


class TestObjectDataDownload(TestWithConnector, TestWithDownloadHandler):
    def setUp(self) -> None:
        TestWithConnector.setUp(self)
        TestWithDownloadHandler.setUp(self)
        self.setup_download_handler(TEST_DATA)
        self.metadata = ObjectMetadata(
            environment=self.environment,
            id=OBJECT_ID,
            name="object.json",
            created_by=...,
            created_at=...,
            modified_at=...,
            modified_by=...,
            parent="/some/path",
            schema_id=...,
            version_id=VERSION_ID,
            stage=None,
        )
        self.download = ObjectDataDownload(
            connector=self.connector, metadata=self.metadata, name=DATA_NAME, initial_url=INITIAL_URL
        )

    def test_label(self) -> None:
        expected = f"{self.metadata.path} (ref={DATA_NAME})"
        self.assertEqual(expected, self.download.label)

    def test_metadata(self) -> None:
        self.assertEqual(self.metadata, self.download.metadata)

    def test_get_cache_location(self) -> None:
        expected = self.cache.get_location(self.environment, "geoscience-object") / str(DATA_NAME)
        self.assertEqual(expected, self.download._get_cache_location(self.cache))

    async def test_get_download_url(self) -> None:
        # Test the initial URL is used first.
        first = await self.download.get_download_url()
        self.assertEqual(INITIAL_URL, first)

        # No requests should be made when using the initial URL.
        self.transport.assert_no_requests()

        # Test that a new URL is generated when the initial URL is used up.
        get_object_response = load_test_data("get_object.json")
        with self.transport.set_http_response(
            status_code=200, content=json.dumps(get_object_response), headers={"Content-Type": "application/json"}
        ):
            second = await self.download.get_download_url()
        self.assertEqual(get_object_response["links"]["data"][1]["download_url"], second)

    async def test_download_to_path(self) -> None:
        dest = self.cache.root / "test_download_to_path.parquet"
        await self.download.download_to_path(dest, self.transport)

        # Check the data was downloaded correctly.
        self.assertEqual(TEST_DATA, dest.read_bytes())
        self.assert_download_requests(INITIAL_URL)

    async def test_download_to_cache(self) -> None:
        dest = await self.download.download_to_cache(self.cache, self.transport)

        # Check the download location is correct.
        expected = self.cache.get_location(self.environment, "geoscience-object") / str(DATA_NAME)
        self.assertEqual(expected, dest)

        # Check the data was downloaded correctly.
        self.assertEqual(TEST_DATA, dest.read_bytes())
        self.assert_download_requests(INITIAL_URL)


class TestObjectDataUpload(TestWithConnector, TestWithUploadHandler):
    def setUp(self) -> None:
        TestWithConnector.setUp(self)
        TestWithUploadHandler.setUp(self)
        self.upload = ObjectDataUpload(
            connector=self.connector,
            environment=self.environment,
            name=DATA_NAME,
            initial_url=INITIAL_URL,
        )

    def test_label(self) -> None:
        expected = DATA_NAME
        self.assertEqual(expected, self.upload.label)

    async def test_get_upload_url(self) -> None:
        # Test the initial URL is used first.
        first = await self.upload.get_upload_url()
        self.assertEqual(INITIAL_URL, first)

        # No requests should be made when using the initial URL.
        self.transport.assert_no_requests()

        # Test that a new URL is generated when the initial URL is used up.
        put_data_response = load_test_data("put_data.json")
        with self.transport.set_http_response(
            status_code=200, content=json.dumps(put_data_response), headers={"Content-Type": "application/json"}
        ):
            second = await self.upload.get_upload_url()
        self.assertEqual(put_data_response[0]["upload_url"], second)

    async def test_upload_from_path(self) -> None:
        source = self.cache.root / "test_upload_from_path.parquet"
        source.write_bytes(TEST_DATA)
        await self.upload.upload_from_path(source, self.transport)

        # Check the file was uploaded correctly.
        uploaded = await self.handler.get_committed()
        self.assertEqual(TEST_DATA, uploaded)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import contextlib
from collections.abc import AsyncGenerator, Iterator, Sequence
from typing import Any, TypeVar
from uuid import UUID

from pydantic import ConfigDict, TypeAdapter

from evo import jmespath, logging
from evo.common import APIConnector, ICache, IFeedback
from evo.common.io.exceptions import DataNotFoundError
from evo.common.utils import NoFeedback, PartialFeedback

from ..data import ObjectMetadata, ObjectReference, ObjectSchema
from ..endpoints import ObjectsApi, models
from ..io import ObjectDataDownload
from . import parse

try:
    import pyarrow as pa
    import pyarrow.compute as pc

    from ..parquet import (
        AttributeInfo,
        CategoryInfo,
        ParquetDownloader,
        ParquetLoader,
        TableInfo,
    )
except ImportError as e:
    print(e)
    _LOADER_AVAILABLE = False
else:
    _LOADER_AVAILABLE = True

    _TABLE_INFO_VALIDATOR: TypeAdapter[TableInfo] = TypeAdapter(TableInfo, config=ConfigDict(extra="ignore"))
    _CATEGORY_INFO_VALIDATOR: TypeAdapter[CategoryInfo] = TypeAdapter(CategoryInfo)
    _NAN_VALIDATOR: TypeAdapter[list[int] | list[float]] = TypeAdapter(
        list[int] | list[float], config=ConfigDict(extra="ignore")
    )
    _ATTRIBUTE_VALIDATOR: TypeAdapter[AttributeInfo] = TypeAdapter(AttributeInfo)

try:
    import pandas as pd
except ImportError:
    _PD_AVAILABLE = False
else:
    _PD_AVAILABLE = True

try:
    import numpy as np
except ImportError:
    _NP_AVAILABLE = False
else:
    _NP_AVAILABLE = True

__all__ = ["DownloadedObject"]

logger = logging.getLogger("object.client")

_T = TypeVar("_T")


def _split_feedback(left: int, right: int) -> float:
    """Helper to split feedback range into two parts based on left and right sizes.

    :param left: Number of parts for the left side of the split.
    :param right: Number of parts for the right side of the split.

    :return: Proportion of feedback to allocate to the left side (between 0.0 and 1.0).
        The right side will be the remainder (1.0 - proportion).

    :raises ValueError: If left or right is negative.
    """
    if left < 0 or right < 0:
        raise ValueError("Left and right sizes must be non-negative")
    elif left >= 0 and right == 0:
        return 1.0  # Left gets all feedback if right is zero
    elif right > 0 and left == 0:
        return 0.0  # Right gets all feedback if left is zero
    else:
        return left / (left + right)  # Proportion of feedback for left


class DownloadedObject:
    """A downloaded geoscience object."""

    def __init__(
        self,
        object_: models.GeoscienceObject,
        metadata: ObjectMetadata,
        urls_by_name: dict[str, str],
        connector: APIConnector,
        cache: ICache | None = None,
    ) -> None:
        """
        :param object_: The raw geoscience object model.
        :param metadata: The parsed metadata for the object.
        :param urls_by_name: A mapping of data names to their initial download URLs.
        :param connector: The API connector to use for downloading data.
        :param cache: An optional cache to use for data downloads.
        """
        self._object = object_
        self._metadata = metadata
        self._urls_by_name = urls_by_name
        self._connector = connector
        self._cache = cache

    @staticmethod
    async def from_reference(
        connector: APIConnector,
        reference: ObjectReference | str,
        cache: ICache | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> DownloadedObject:
        """Download a geoscience object from the service, given an object reference.

        :param connector: The API connector to use for downloading data.
        :param reference: The reference to the object to download, or a URL as a string that can be parsed into
            a reference.
        :param cache: An optional cache to use for data downloads.
        :param request_timeout: An optional timeout to use for API requests. See evo.common.APIConnector for details.

        :raises ValueError: If the reference is invalid, or if the connector base URL does not match the reference hub URL.
        """
        ref = ObjectReference(reference)  # Parse the reference if it's a string

        if connector.base_url != ref.hub_url:
            raise ValueError(
                f"The connector base URL '{connector.base_url}' does not match the reference hub URL '{ref.hub_url}'"
            )

        api = ObjectsApi(connector)

        request_kwargs = dict(
            org_id=str(ref.org_id),
            workspace_id=str(ref.workspace_id),
            version=ref.version_id,
            additional_headers={"Accept-Encoding": "gzip"},
            request_timeout=request_timeout,
        )

        if ref.object_id is not None and ref.object_path is not None:
            raise ValueError("Only one of object_id or object_path should be provided")

        if ref.object_id is not None:
            response = await api.get_object_by_id(object_id=ref.object_id, **request_kwargs)
        elif ref.object_path is not None:
            response = await api.get_object(objects_path=ref.object_path, **request_kwargs)
        else:
            raise ValueError("Either object_id or object_path must be provided")

        metadata = parse.object_metadata(response, ref.environment)
        urls_by_name = {getattr(link, "name", link.id): link.download_url for link in response.links.data}
        return DownloadedObject(
            object_=response.object,
            metadata=metadata,
            urls_by_name=urls_by_name,
            connector=connector,
            cache=cache,
        )

    @property
    def schema(self) -> ObjectSchema:
        """The schema of the object."""
        return self._metadata.schema_id

    @property
    def metadata(self) -> ObjectMetadata:
        """The metadata of the object."""
        return self._metadata

    def as_dict(self) -> dict:
        """Get this object as a dictionary."""
        return self._object.model_dump(mode="python", by_alias=True)

    def search(self, expression: str) -> Any:
        """Search the object metadata using a JMESPath expression.

        :param expression: The JMESPath expression to use for the search.

        :return: The result of the search.
        """
        return jmespath.search(expression, self.as_dict())

    def prepare_data_download(self, data_identifiers: Sequence[str | UUID]) -> Iterator[ObjectDataDownload]:
        """Prepare to download multiple data files from the geoscience object service, for this object.

        Any data IDs that are not associated with the requested object will raise a DataNotFoundError.

        :param data_identifiers: A list of sha256 digests or UUIDs for the data to be downloaded.

        :return: An iterator of data download contexts that can be used to download the data.

        :raises DataNotFoundError: If any requested data ID is not associated with this object.
        """
        try:
            filtered_urls_by_name = {str(name): self._urls_by_name[str(name)] for name in data_identifiers}
        except KeyError as exc:
            raise DataNotFoundError(f"Unable to find the requested data: {exc.args[0]}") from exc
        for ctx in ObjectDataDownload._create_multiple(
            connector=self._connector, metadata=self._metadata, urls_by_name=filtered_urls_by_name
        ):
            yield ctx

    async def update(
        self,
        object_dict: dict,
        check_for_conflict: bool = True,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> DownloadedObject:
        """Update the geoscience object on the geoscience object service. Returning a new DownloadedObject representing
        the new version of the object.

        This will create a new version of the object, that fully replaces the existing properties of the object with
        those provided in `object_dict`.

        Note, this will not update the "DownloadedObject" instance in-place - it will still represent the original
        version of the object. You will need to download the updated version separately if you wish to work with it.

        :param object_dict: The new properties of the object as a dictionary.
        :param check_for_conflict: If True, and if a newer version of the object exists on the geoscience object
            service, the update will fail with a ObjectModifiedError exception. If False, it will not check whether
            there is a newer version, so will perform the update regardless.
        :param request_timeout: An optional timeout to use for API requests. See evo.common.APIConnector for details.

        :returns: The new version of the object as a DownloadedObject.
        """

        api = ObjectsApi(self._connector)

        if "uuid" not in object_dict:
            object_dict["uuid"] = self.metadata.id

        model = models.UpdateGeoscienceObject.model_validate(object_dict)
        if model.uuid != self.metadata.id:
            raise ValueError("The object ID in the new object does not match the current object ID")

        response = await api.update_objects_by_id(
            object_id=str(self.metadata.id),
            org_id=str(self.metadata.environment.org_id),
            workspace_id=str(self.metadata.environment.workspace_id),
            update_geoscience_object=model,
            request_timeout=request_timeout,
            if_match=self.metadata.version_id if check_for_conflict else None,
        )
        metadata = parse.object_metadata(response, self.metadata.environment)
        urls_by_name = {getattr(link, "name", link.id): link.download_url for link in response.links.data}
        return DownloadedObject(
            object_=response.object,
            metadata=metadata,
            urls_by_name=urls_by_name,
            connector=self._connector,
            cache=self._cache,
        )

    if _LOADER_AVAILABLE:
        # Optional support for loading Parquet data using PyArrow.

        def _validate_typed_dict(self, value: _T | str, validator: TypeAdapter[_T]) -> _T:
            if isinstance(value, str):
                resolved = self.search(value)
                # Implicitly unwrap single-element arrays for convenience
                # This allows, using predicates like: attributes[?name=='my_attribute']
                if isinstance(resolved, jmespath.JMESPathArrayProxy) and len(resolved) == 1:
                    resolved = resolved[0]
                if isinstance(resolved, jmespath.JMESPathObjectProxy):
                    value = resolved.raw
                else:
                    raise ValueError(f"Expected object, got {type(resolved)}")
            return validator.validate_python(value)

        def _validate_nan_values(self, nan_values: list[int] | list[float] | str | None) -> list[int] | list[float]:
            if nan_values is None:
                return []
            if isinstance(nan_values, str):
                resolved = self.search(nan_values)
                if isinstance(resolved, jmespath.JMESPathArrayProxy) and len(resolved) == 1:
                    # Consider single-element arrays for unwrapping
                    # This allows, using predicates like: attributes[?name=='my_attribute']
                    child = resolved[0]
                    if not isinstance(child, (int, float)):
                        resolved = child
                if isinstance(resolved, jmespath.JMESPathArrayProxy):
                    nan_values = resolved.raw
                # Support passing nan_description structure too
                elif isinstance(resolved, jmespath.JMESPathObjectProxy) and "values" in resolved.raw:
                    nan_values = resolved.raw["values"]
                else:
                    raise ValueError(f"Expected list, got {type(resolved)}")
            return _NAN_VALIDATOR.validate_python(nan_values)

        @contextlib.asynccontextmanager
        async def _with_parquet_loader(
            self, table_info: TableInfo | str, fb: IFeedback
        ) -> AsyncGenerator[ParquetLoader, None]:
            """Download parquet data and get a ParquetLoader for the data referenced by the given
            table info or data reference string.

            :param table_info: The table info dict, JMESPath to table info within the object.
            :param fb: An optional feedback instance to report download progress to.

            :returns: A ParquetLoader that can be used to read the referenced data.
            """
            table_info = self._validate_typed_dict(table_info, _TABLE_INFO_VALIDATOR)
            (download,) = self.prepare_data_download([table_info["data"]])
            async with ParquetDownloader(download, self._connector.transport, self._cache).with_feedback(fb) as loader:
                loader.validate_with_table_info(table_info)
                yield loader

        async def download_table(
            self,
            table_info: TableInfo | str,
            fb: IFeedback = NoFeedback,
            *,
            nan_values: list[int] | list[float] | str | None = None,
            column_names: Sequence[str] | None = None,
        ) -> pa.Table:
            """Download the data referenced by the given table info as a PyArrow Table.

            :param table_info: The table info dict, ot JMESPath to table info within the object.
            :param fb: An optional feedback instance to report download progress to.
            :param nan_values: An optional list of values to treat as null. Can also be a JMESPath expression to the
               list of nan values, or the nan_description structure.
            :param column_names: An optional list of column names for the table, instead of those in the Parquet file.

            :returns: A PyArrow Table containing the downloaded data.
            """
            async with self._with_parquet_loader(table_info, fb) as loader:
                table = loader.load_as_table()

            if column_names is None:
                column_names = table.column_names

            nan_values = self._validate_nan_values(nan_values)
            if len(nan_values) == 0:
                return table.rename_columns(column_names)

            # Replace specified nan_values with nulls
            arrays = []
            for array in table.columns:
                if isinstance(array, pa.ChunkedArray):
                    array = array.combine_chunks()
                null_scalar = pa.scalar(None, type=array.type)
                nan_value_array = pa.array(nan_values, type=array.type)
                arrays.append(pc.replace_with_mask(array, pc.is_in(array, nan_value_array), null_scalar))
            return pa.Table.from_arrays(arrays, names=column_names)

        async def download_category_table(
            self,
            category_info: CategoryInfo | str,
            *,
            nan_values: list[int] | list[float] | str | None = None,
            column_names: Sequence[str] | None = None,
            fb: IFeedback = NoFeedback,
        ) -> pa.Table:
            """Download the data referenced by the given category info as a PyArrow Table.

            The arrays into the table will be DictionaryArrays constructed from the values and lookup tables.

            :param category_info: The category info dict, or JMESPath to the category info within the object.
            :param nan_values: An optional list of values to treat as null. Can also be a JMESPath expression to
                nan_description structure.
            :param column_names: An optional list of column names for the table, instead of those in the Parquet file.
            :param fb: An optional feedback instance to report download progress to.

            :returns: A PyArrow Table containing the downloaded data.
            """
            category_info = self._validate_typed_dict(category_info, _CATEGORY_INFO_VALIDATOR)

            v_size = (
                category_info["values"]["length"] * category_info["values"]["width"]
            )  # Total number of cells in values
            t_size = category_info["table"]["length"] * 2  # Lookup tables always have 2 columns
            split = _split_feedback(v_size, t_size)

            values_table = await self.download_table(
                category_info["values"],
                nan_values=nan_values,
                column_names=column_names,
                fb=PartialFeedback(fb, start=0, end=split),
            )
            lookup_table = await self.download_table(category_info["table"], fb=PartialFeedback(fb, start=split, end=1))

            arrays = []
            for array in values_table.columns:
                indices = pc.index_in(array, lookup_table[0])
                arrays.append(pa.DictionaryArray.from_arrays(indices, lookup_table[1]))
            return pa.Table.from_arrays(arrays, names=values_table.column_names)

        async def download_attribute_table(
            self,
            attribute: AttributeInfo | str,
            fb: IFeedback = NoFeedback,
        ) -> pa.Table:
            """Download the data referenced by the given attribute as a PyArrow Table.

            :param attribute: The attribute info dict, or JMESPath to the attribute info within the object.
            :param fb: An optional feedback instance to report download progress to.

            :returns: A PyArrow Table containing the downloaded data.
            """
            attribute = self._validate_typed_dict(attribute, _ATTRIBUTE_VALIDATOR)

            if "table" in attribute:
                table = await self.download_category_table(
                    attribute,
                    nan_values=attribute["nan_description"]["values"] if "nan_description" in attribute else None,
                    fb=fb,
                )
            else:
                table = await self.download_table(
                    attribute["values"],
                    nan_values=attribute["nan_description"]["values"] if "nan_description" in attribute else None,
                    fb=fb,
                )
            if len(table.column_names) == 1:
                table = table.rename_columns([attribute["name"]])
            else:
                table = table.rename_columns([f"{attribute['name']}_{i}" for i in range(len(table.column_names))])
            return table

        if _PD_AVAILABLE:
            # Optional support for loading data as Pandas DataFrames. Requires parquet support via PyArrow as well.

            async def download_dataframe(
                self,
                table_info: TableInfo | str,
                fb: IFeedback = NoFeedback,
                *,
                nan_values: list[int] | list[float] | str | None = None,
                column_names: Sequence[str] | None = None,
            ) -> pd.DataFrame:
                """Download the data referenced by the given table info as a Pandas DataFrame.

                :param table_info: The table info dict, JMESPath to table info within the object.
                :param fb: An optional feedback instance to report download progress to.
                :param nan_values: An optional list of values to treat as null. Can also be a JMESPath expression to
                    nan_description structure.
                :param column_names: An optional list of column names for the table, instead of those from the Parquet file.

                :returns: A Pandas DataFrame containing the downloaded data.
                """
                table = await self.download_table(table_info, fb=fb, nan_values=nan_values, column_names=column_names)
                return table.to_pandas()

            async def download_category_dataframe(
                self,
                category_info: CategoryInfo | str,
                fb: IFeedback = NoFeedback,
                *,
                nan_values: list[int] | list[float] | str | None = None,
                column_names: Sequence[str] | None = None,
            ) -> pd.DataFrame:
                """Download the data referenced by the given category info as a Pandas DataFrame.

                :param category_info: The category info dict, or JMESPath to the category info within the object.
                :param nan_values: An optional list of values to treat as null. Can also be a JMESPath expression to
                    nan_description structure.
                :param column_names: An optional list of column names for the table, instead of those from the Parquet file.
                :param fb: An optional feedback instance to report download progress to.

                :returns: A Pandas DataFrame containing the downloaded data.
                """
                table = await self.download_category_table(
                    category_info, fb=fb, nan_values=nan_values, column_names=column_names
                )
                return table.to_pandas()

            async def download_attribute_dataframe(
                self,
                attribute: AttributeInfo | str,
                fb: IFeedback = NoFeedback,
            ) -> pd.DataFrame:
                """Download the data referenced by the given attribute as a Pandas DataFrame.

                :param attribute: The attribute info dict, or JMESPath to the attribute within the object.
                :param fb: An optional feedback instance to report download progress to.

                :returns: A Pandas DataFrame containing the downloaded data.
                """
                attribute = self._validate_typed_dict(attribute, _ATTRIBUTE_VALIDATOR)

                if "table" in attribute:
                    df = await self.download_category_dataframe(
                        attribute,
                        nan_values=attribute["nan_description"]["values"] if "nan_description" in attribute else None,
                        fb=fb,
                    )
                else:
                    df = await self.download_dataframe(
                        attribute["values"],
                        nan_values=attribute["nan_description"]["values"] if "nan_description" in attribute else None,
                        fb=fb,
                    )
                if len(df.columns) == 1:
                    df.columns = [attribute["name"]]
                else:
                    df.columns = [f"{attribute['name']}_{i}" for i in range(len(df.columns))]
                return df

        if _NP_AVAILABLE:
            # Optional support for loading data as NumPy arrays. Requires parquet support via PyArrow as well.

            async def download_array(self, table_info: TableInfo | str, fb: IFeedback = NoFeedback) -> np.ndarray:
                """Download the data referenced by the given table info as a NumPy array.

                :param table_info: The table info dict, JMESPath to table info within the object.
                :param fb: An optional feedback instance to report download progress to.

                :returns: A NumPy array containing the downloaded data.
                """
                async with self._with_parquet_loader(table_info, fb) as loader:
                    return loader.load_as_array()



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Compute Task API
=============

Normally this code would be generated from an OpenAPI specification. However, the API specification for
the Compute Task API is new and not yet finalized. Therefore, this code is written against the intended API,
with some customisation as needed to make it work.

The client implementation should still abstract these details away from the user, so that the user can interact
with the API in a more user-friendly way, and so that this implementation can be changed without affecting the user.

This code is based on the OpenAPI specification for Compute Task API.
API version: 0.0.1
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, HTTPResponse, RequestMethod

from ..models import *  # noqa: F403

__all__ = ["JobApi"]


class JobApi:
    """API client for the job endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def cancel_job(
        self,
        org_id: str,
        topic: str,
        task: str,
        job_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> EmptyResponse:
        """Cancels a specific job

        Cancel a specific job, identified by the organization ID, topic, task, and job ID. The job's status is updated to 'canceling' and the job is stopped.

        :param org_id: The organization identifier
            Format: `uuid`
            Example: `'org_id_example'`
        :param topic: The compute topic
            Example: `'topic_example'`
        :param task: The compute task
            Example: `'task_example'`
        :param job_id: The job identifier
            Format: `uuid`
            Example: `'job_id_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """  # noqa: E501
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "topic": topic,
            "task": task,
            "job_id": job_id,
        }

        # Prepare the header parameters.
        _header_params = {}
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "204": EmptyResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.DELETE,
            resource_path="/compute/orgs/{org_id}/{topic}/{task}/{job_id}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_job_results(
        self,
        org_id: str,
        topic: str,
        task: str,
        job_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> HTTPResponse:  # noqa: F405
        """Retrieves the results of a specific job

        Get the results of a specific job, identified by the organization ID, topic, task, and job ID. The job's status is returned in the response, along with either the results or an error object, depending on the status of the job. If the job is ongoing, the response includes links to the job's cancellation endpoint and detailed status endpoint.

        :param org_id: The organization identifier
            Format: `uuid`
            Example: `'org_id_example'`
        :param topic: The compute topic
            Example: `'topic_example'`
        :param task: The compute task
            Example: `'task_example'`
        :param job_id: The job identifier
            Format: `uuid`
            Example: `'job_id_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """  # noqa: E501
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "topic": topic,
            "task": task,
            "job_id": job_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            # Get the raw response so we can include headers if the job failed.
            "200": HTTPResponse,
            "202": HTTPResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/compute/orgs/{org_id}/{topic}/{task}/{job_id}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_job_status(
        self,
        org_id: str,
        topic: str,
        task: str,
        job_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> JobStatusResponse:  # noqa: F405
        """Retrieves the status of a specific job

        Get the status of a specific job, identified by the organization ID, topic, task, and job ID. The job's status is returned in the response, along with a progress indicator, a message, and an error object, if applicable. If the job is ongoing, the response includes links to the job's cancellation endpoint.

        :param org_id: The organization identifier
            Format: `uuid`
            Example: `'org_id_example'`
        :param topic: The compute topic
            Example: `'topic_example'`
        :param task: The compute task
            Example: `'task_example'`
        :param job_id: The job identifier
            Format: `uuid`
            Example: `'job_id_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """  # noqa: E501
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
            "topic": topic,
            "task": task,
            "job_id": job_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        }
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": JobStatusResponse,  # noqa: F405
            "202": JobStatusResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/compute/orgs/{org_id}/{topic}/{task}/{job_id}/status",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )
